
#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.transpose_code_chunk import TransposeCodeChunk

from mitosheet.state import DATAFRAME_SOURCE_TRANSPOSED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID
from mitosheet.utils import get_first_unused_dataframe_name

class TransposeStepPerformer(StepPerformer):
    """
    Allows you to transpose.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'transpose'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')

        # We make a new state to modify it
        post_state = prev_state.copy()

        pandas_start_time = perf_counter()
        new_df = post_state.dfs[sheet_index].T
        pandas_processing_time = perf_counter() - pandas_start_time

        new_df_name = get_first_unused_dataframe_name(post_state.df_names, f'{post_state.df_names[sheet_index]}_transposed')
        post_state.add_df_to_state(new_df, DATAFRAME_SOURCE_TRANSPOSED, df_name=new_df_name)

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            TransposeCodeChunk(prev_state, post_state, get_param(params, 'sheet_index'))
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

"""
Utilities for step performers
"""

from typing import Any, Dict, List

import pandas as pd
from mitosheet.column_headers import try_make_new_header_valid_if_multi_index_headers

from mitosheet.types import ColumnHeader


def get_param(params: Dict[str, Any], key: str) -> Any:
    if key in params:
        return params[key]
    return None

def add_columns_to_df(df: pd.DataFrame, new_columns_df: pd.DataFrame, new_column_headers: List[ColumnHeader], column_index: int) -> pd.DataFrame:
    """
    Adds new_columns_df to df at column_index.
    """
    # Make sure the new column headers are valid before adding them to the dataframe
    df_column_headers = df.columns
    new_column_headers = [try_make_new_header_valid_if_multi_index_headers(df_column_headers, column_header) for column_header in new_column_headers]
    # Add the new columns to the end of the dataframe
    df[new_column_headers] = new_columns_df
    # Set the columns in the correct order
    df = df[df.columns[:column_index + 1].tolist() + new_column_headers + df.columns[column_index + 1:-len(new_column_headers)].tolist()]
    return df


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy, deepcopy
from typing import Any, Dict, List, Optional, Set, Tuple
import uuid
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import GraphID


class GraphDuplicateStepPerformer(StepPerformer):
    """
    This steps duplicates a graph of a given graphID. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'graph_duplicate'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        old_graph_id: GraphID = get_param(params, 'old_graph_id')
        new_graph_id: GraphID = get_param(params, 'new_graph_id')

        post_state = prev_state.copy()

        # Execute the step
        graph_copy = deepcopy(post_state.graph_data_dict[old_graph_id])
        # We don't need to insist the the graph names are unique because they are just used in 
        # the sheet tab display. They aren't used in generated code or to identify graphs in the steps
        graph_copy["graphTabName"] = graph_copy["graphTabName"] + '_copy'
        
        # Add the duplicated graph to the graph_data
        post_state.graph_data_dict[new_graph_id] = graph_copy
        
        return post_state, {
            'pandas_processing_time': 0 # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:

        # Graph steps don't add any generated code to the analysis script. 
        return [
            EmptyCodeChunk(
                prev_state, 
                post_state, 
                'Duplicated graph',
                'Duplicated a graph',
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import deepcopy
from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.column_steps.add_column_code_chunk import AddColumnCodeChunk

from mitosheet.errors import make_column_exists_error, make_no_sheet_error
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param


class AddColumnStepPerformer(StepPerformer):
    """""
    A add_column step, which allows you to add a column to 
    a dataframe. 
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'add_column'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_header: str = get_param(params, 'column_header')
        column_header_index: int = get_param(params, 'column_header_index')
            
        # if the sheet doesn't exist, throw an error
        if not prev_state.does_sheet_index_exist_within_state(sheet_index):
            raise make_no_sheet_error({sheet_index})

        if column_header in prev_state.dfs[sheet_index].keys():
            raise make_column_exists_error(column_header)

        # We add a new step with the added column
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        # If the column_header_index is out of range, then make the new column the last column
        if column_header_index < 0 or len(prev_state.dfs[sheet_index].columns) <= column_header_index:
            column_header_index = len(prev_state.dfs[sheet_index].columns)

        # Update the state variables
        post_state.add_columns_to_state(sheet_index, [column_header])
            
        # Update the dataframe
        pandas_start_time = perf_counter()
        post_state.dfs[sheet_index].insert(column_header_index, column_header, 0)
        pandas_processing_time = perf_counter() - pandas_start_time
        
        return post_state, {
            'column_header_index': column_header_index,
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            AddColumnCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),    
                get_param(params, 'column_header'),
                execution_data.get('column_header_index', 0) if execution_data is not None else 0  
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
import os
from typing import Any, Dict, List, Optional
import platform
import string

def get_path_modified(path: str, f: str) -> Optional[float]:
    """
    For a path, returns when it was last modified. If that path is unaccessible, 
    for example if the path is a disconnected Google Drive that can no longer be read,
    it returns None
    """
    try:
        return os.path.getmtime(os.path.join(path, f))
    except:
        return None

def get_windows_drives() -> List[str]:
    """
    Returns a list of all the drives on a Windows machine. 
    See https://stackoverflow.com/questions/827371/is-there-a-way-to-list-all-the-available-windows-drives 
    for more information.
    """
    drives = []

    # Ctypes only exports windll on windows computers. 
    # Read more here: https://docs.python.org/3/library/ctypes.html#module-ctypes
    if platform.system() == 'Windows':
        from ctypes import windll # type: ignore
        bitmask = windll.kernel32.GetLogicalDrives()
        for letter in string.ascii_uppercase:
            if bitmask & 1:
                # Add :\ to the end to complete the drive. Not having :\ on the end causes os to fail when reading the drive
                drives.append(letter + ':\\') 
            bitmask >>= 1

    return drives


def get_path_parts(path: str) -> List[str]:
    """
    For a path, returns a list of the path broken down into
    pieces.
    """
    
    # On Windows, drive will be C: or D:, etc. On Unix, drive will be empty.
    drive, path_and_file = os.path.splitdrive(path)
    path, file = os.path.split(path_and_file)

    folders = []
    # https://stackoverflow.com/questions/3167154/how-to-split-a-dos-path-into-its-components-in-python
    # We take the code from here, which apparently avoids entering an infinite loop
    # as it breaks when path != ''
    while 1:
        path, folder = os.path.split(path)

        if folder != "":
            folders.append(folder)
        elif path != "":
            folders.append(path)

            break

    folders.reverse()

    if drive != '':
        # If the drive is not empty, we are on windows, and we need to do some path hackery to get things to
        # work well. First, we add the \ to the start of the path, and then we make sure the drive has a slash
        # after it. Then, we remove the slash from the folder list if it exists at the start
        if len(folders) > 0 and folders[0] == '\\':
            folders = folders[1:]

        return ['\\', drive + '\\'] + folders + ([file] if file != '' else [])
    else:
        # If the drive is '' then we are on a Linux system and the root folder / is contained in the folders.
        # We get rid of the empty path so that we can easily handle windows and linux paths the same.
        return folders + ([file] if file != '' else [])


def get_path_contents(params: Dict[str, Any]) -> str:
    """
    Takes an event with path parts in a list, turns them into
    an actual path, and then sends an API response with those 
    path parts
    """ 
    path_parts = params['path_parts']

    # Join the path and normalize it (note this should be OS independent)
    path = os.path.join(*path_parts)
    path = os.path.normpath(path)

    if path == '\\' and platform.system() == 'Windows':
        # If the path only has one part, it means they are accessing the root folder. If the user is on
        # Windows, this folder doesn't exist so we fake one by letting them pick amongst their drives.
        filenames: List[str] = []
        dirnames = get_windows_drives()
    else:
        # We default the path to "." on the frontend, but we replace
        # this with the current directory full path so we can get all
        # the path parts correctly 
        if path == '.':
            path = os.getcwd()

        try:
            # This loop defines these variables, but does nothing with them
            # so we can then return them (which is why we break immediately).
            # If the path is just a windows drive that is formatted like C or C:, instead of C:/,
            # then os.walk does not work properly. 
            for (dirpath, dirnames, filenames) in os.walk(path):
                break

            # We sort the files so they are alphabetical (ignoring case)
            dirnames = sorted(dirnames, key=str.lower)
            filenames = sorted(filenames, key=str.lower)
        except:
            # If we cannot read the current path, this is a result of the fact
            # that there are permission errors (or something), in which case we
            # just return an empty result
            dirnames = []
            filenames = []
    
    # We then filter out any hidden folders and files, which we don't want users to be able
    # to see (as they would otherwise). 
    # Linux, Max == starts with "."
    # Windows == "$"
    # We also filter out files that start with:
    # ~$ is used to prefix hidden temporary files that are created when a document is opened in Windows
    dirnames = [d for d in dirnames if (not d.startswith('.') and not d.startswith('$'))]
    filenames = [f for f in filenames if (not f.startswith('.') and not f.startswith('$') and not f.startswith('~$'))]

    return json.dumps({
        'path': path,
        'path_parts': get_path_parts(path),
        # For each element, record if it's a directory, and the time it was last modified
        'elements': [
            {'name': f, 'isDirectory': False, 'lastModified': get_path_modified(path, f)} for f in filenames
        ] + [
            {'name': d, 'isDirectory': True, 'lastModified': get_path_modified(path, d)} for d in dirnames
        ]
    })
    
    
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import subprocess
import sys
from typing import Any, Dict

import atheris
import pandas as pd

from mitosheet.mito_backend import get_mito_frontend_code

with atheris.instrument_imports():
    import mitosheet
    from mitosheet.step_performers.filter import FC_NUMBER_EXACTLY
    from mitosheet.tests.test_utils import (MitoWidgetTestWrapper,
                                            create_mito_wrapper_dfs)



def get_random_event(fdp: atheris.FuzzedDataProvider) -> Dict[str, Any]:
    """
    Takes a fuzz data provider, returns a random
    event to send to the backend.
    
    TODO: for now, we just use one dataframe
    """
    return {
        'type': 'add_column_edit',
        'step_id': fdp.ConsumeUnicode(4),
        'id': fdp.ConsumeUnicode(4),
        'params': {
            'sheet_index': 0,
            'column_header': fdp.ConsumeUnicode(100),
            'column_header_index': fdp.ConsumeInt(3)
        }
    }


def send_random_event(mito: MitoWidgetTestWrapper, fdp: atheris.FuzzedDataProvider) -> None:

    random_sheet_index = fdp.PickValueInList(list(range(len(mito.dfs))))
    random_existing_column_header = fdp.PickValueInList(list(mito.dfs[random_sheet_index].keys()))

    other_random_sheet_index = fdp.PickValueInList(list(range(len(mito.dfs))))
    random_column_header = fdp.ConsumeUnicode(1)
    other_random_column_header = fdp.ConsumeUnicode(1)

    event = fdp.ConsumeIntInRange(0, 20)

    print(event)

    if event == 0:
        mito.add_column(
            random_sheet_index,
            random_column_header,
            fdp.ConsumeInt(3)
        )
    elif event == 1:
        mito.set_formula(
            '=0', # TODO: randomize this 
            random_sheet_index,
            random_column_header, 
            False,
        )

    elif event == 2:
        mito.merge_sheets(
            'left', # TODO: randomize this
            random_sheet_index, 
            random_column_header, 
            [random_column_header],
            other_random_sheet_index, 
            other_random_column_header,
            [other_random_column_header]
        )

    elif event == 3:
        mito.drop_duplicates(
            random_sheet_index, 
            [random_existing_column_header], # TODO: randomize this 
            'last', # TODO: randomize this
        )

    elif event == 4:
        mito.pivot_sheet(
            random_sheet_index, 
            # TODO: take these from the pivoted sheet
            [random_column_header],
            [other_random_column_header],
            {random_column_header: ['sum']},
            flatten_column_headers=True,
            # TODO: add other params here
        )

    elif event == 5:
        mito.filter(
            random_sheet_index,
            random_column_header,
            'And',
            FC_NUMBER_EXACTLY,
            1
        )
    elif event == 6:
        mito.sort(
            random_sheet_index, 
            random_column_header,
            'ascending'
        )

    elif event == 7:
        mito.reorder_column(
            random_sheet_index, 
            random_column_header, 
            fdp.ConsumeInt(3)
        )

    elif event == 8:
        mito.rename_column(
            random_sheet_index, 
            random_column_header, 
            other_random_column_header
        )

    elif event == 9:
        mito.delete_columns(
            random_sheet_index,
            [random_column_header]
        )

    elif event == 10:
        mito.change_column_dtype(
            random_sheet_index, 
            random_column_header, 
            'string'
        )

    elif event == 11:
        mito.bulk_old_rename(
            move_to_deprecated_id_algorithm=fdp.ConsumeBool()
        )

    elif event == 12:
        mito.undo()

    elif event == 13:
        mito.redo()

    elif event == 14:
        mito.clear()

    elif event == 15:
        mito.delete_dataframe(random_sheet_index)

    elif event == 16:
        mito.duplicate_dataframe(random_sheet_index)

    elif event == 17:
        mito.rename_dataframe(random_sheet_index, random_column_header)

    elif event == 18:
        mito.set_cell_value(
            random_sheet_index, 
            random_column_header,
            fdp.ConsumeInt(3),
            other_random_column_header
        )

    elif event == 19:
        pass # TODO: simple import
    elif event == 20:
        pass # TODO: excel import


NUM_EVENTS = 20

def TestMito(data):
    import random
    fdp = atheris.FuzzedDataProvider(data)

    df = pd.DataFrame({'A': []})
    mito = create_mito_wrapper_dfs(df)
    frontend_code = get_mito_frontend_code('1', '2', '3', mito.mito_backend)
    
    file = 'tests/out.js'
    with open(file, 'w+') as f:
        f.write(frontend_code)

    p = subprocess.Popen(['node', file], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (_, err) = p.communicate()
    p.wait()

    # we want to make sure that there are no failures in parsing, that it runs up to the 
    # ReferenceError: document is not defined 
    assert 'SyntaxError' not in err.decode('utf-8') 
    assert 'ReferenceError' in err.decode('utf-8')

    print("NUM STEPS:", len(mito.steps_including_skipped))


atheris.Setup(sys.argv, TestMito)
atheris.Fuzz()



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains helpful utility functions
"""
import json
from random import randint
import re
import uuid
from typing import Any, Dict, List, Optional, Set, Tuple
import os

import numpy as np
import pandas as pd

from mitosheet.column_headers import ColumnIDMap, get_column_header_display
from mitosheet.sheet_functions.types.utils import get_float_dt_td_columns
from mitosheet.types import (ColumnHeader, ColumnID, DataframeFormat, StateType, FrontendFormula)

# We only send the first 1500 rows of a dataframe; note that this
# must match this variable defined on the front-end
MAX_ROWS = 1_500
MAX_COLUMNS = 1_500

def get_first_unused_dataframe_name(existing_df_names: List[str], new_dataframe_name: str) -> str:
    """
    Appends _1, _2, .. to df name until it finds an unused 
    dataframe name. If no append is necessary, will just
    return the initial passed value.
    """
    if new_dataframe_name not in existing_df_names:
        return new_dataframe_name

    for i in range(len(existing_df_names) + 1):
        new_name = f'{new_dataframe_name}_{i + 1}'
        if new_name not in existing_df_names:
            return new_name

    # NOTE: We should never reach this, as the above loop will
    # find a dataframe name that is taken
    return new_dataframe_name

def get_valid_dataframe_name(existing_df_names: List[str], original_dataframe_name: str) -> str:
    """
    Given a string, turns it into a valid dataframe name, making sure to 
    not overlap with any existing dataframe names.
    """
    # We get all the words from the original name, and append them with underscores
    dataframe_name = '_'.join(filter(None, [
        match.group() if match.group() != 'csv' and match.group() != 'xlsx' else '' for match in re.finditer('\w+', original_dataframe_name)
    ]))

    # A valid variable name cannot be empty, or start with a number
    if len(dataframe_name) == 0 or dataframe_name[0].isdecimal():
        return get_first_unused_dataframe_name(existing_df_names, 'df_' + dataframe_name)

    return get_first_unused_dataframe_name(existing_df_names, dataframe_name)


def get_valid_dataframe_names(existing_df_names: List[str], original_df_names: List[str]) -> List[str]:
    """
    Helper function for taking a list of potential and turning them into valid
    names for dataframes, that do not overlap with the existing dataframe names.
    """

    final_names: List[str] = []

    for original_df_name in original_df_names:
        new_names_final = get_valid_dataframe_name(
            existing_df_names + final_names,
            original_df_name
        )
        final_names.append(new_names_final)
    
    return final_names

def is_default_df_names(df_names: List[str]) -> bool:
    """
    Returns true if the df names list is the default df names
    created when the widget state container is initialized
    """
    return len(df_names) > 0 and df_names == [f'df{i + 1}' for i in range(len(df_names))]


def dfs_to_array_for_json(
        state: StateType,
        modified_sheet_indexes: Set[int],
        previous_array: List,
        dfs: List[pd.DataFrame],
        df_names: List[str],
        df_sources: List[str],
        column_formulas_array: List[Dict[ColumnID, FrontendFormula]],
        column_filters_array: List[Dict[ColumnID, Any]],
        column_ids: ColumnIDMap,
        df_formats: List[DataframeFormat]
    ) -> List:

    new_array = []
    for sheet_index, df in enumerate(dfs):
        if sheet_index in modified_sheet_indexes:
            new_array.append(
                df_to_json_dumpsable(
                    state,
                    df, 
                    sheet_index,
                    df_names[sheet_index],
                    df_sources[sheet_index],
                    column_formulas_array[sheet_index],
                    column_filters_array[sheet_index],
                    column_ids.column_header_to_column_id[sheet_index],
                    df_formats[sheet_index],
                    # We only send the first 1500 rows and 1500 columns
                    max_rows=MAX_ROWS,
                    max_columns=MAX_COLUMNS
                ) 
            )
        else:
            new_array.append(previous_array[sheet_index])

    return new_array


def df_to_json_dumpsable(
        state: StateType,
        original_df: pd.DataFrame,
        sheet_index: int,
        df_name: str,
        df_source: str,
        column_formulas: Dict[ColumnID, FrontendFormula],
        column_filters: Dict[ColumnID, Any],
        column_headers_to_column_ids: Dict[ColumnHeader, ColumnID],
        df_format: DataframeFormat,
        max_rows: Optional[int]=MAX_ROWS, # How many items you want to display. None when using this function to get unique value counts
        max_columns: int=MAX_COLUMNS # How many columns you want to display. Unlike max_rows, this is always defined
    ) -> Dict[str, Any]:
    """
    Returns a dataframe and other metadata represented in a way that can be turned into a 
    JSON object with json.dumps.

    Should follow the format:
    {
        dfName: string;
        dfSource: DFSource;
        numRows: number,
        numColumns: number,
        data: {
            columnID: string;
            columnHeader: (string | number);
            columnDtype: string;
            columnData: (string | number)[];
        }[];
        columnIDsMap: ColumnIDsMap;
        columnSpreadsheetCodeMap: Record<string, string>;
        columnFiltersMap: ColumnFilterMap;
        columnnDtypeMap: Record<ColumnID, string>;
        index: (string | number)[];
        df_format: DataframeFormat;
        conditionalFormattingResult: ConditionalFormattingResult
    }
    """

    (num_rows, num_columns) = original_df.shape 

    json_obj = convert_df_to_parsed_json(original_df, max_rows=max_rows, max_columns=max_columns)

    final_data = []
    column_dtype_map = {}
    for column_index, column_header in enumerate(original_df.columns):
        column_id = column_headers_to_column_ids[column_header]

        column_final_data: Dict[str, Any] = {
            'columnID': column_id,
            'columnHeader': get_column_header_display(column_header),
            'columnDtype': str(original_df[column_header].dtype),
            'columnData': [],
        }
        column_dtype_map[column_id] = str(original_df[column_header].dtype)
        for row in json_obj['data']:
            # If we're beyond the max columns, we might not have data, and we leave column data empty
            # in this case and don't append anything
            column_final_data['columnData'].append(row[column_index] if column_index < MAX_COLUMNS else None)
        
        final_data.append(column_final_data) 

    # Import just before we use it to avoid circular imports
    from mitosheet.pro.conditional_formatting_utils import get_conditonal_formatting_result
    
    return {
        "dfName": df_name,
        "dfSource": df_source,
        'numRows': num_rows,
        'numColumns': num_columns,
        'data': final_data,
        # NOTE: We make sure that all the maps are in the correct order, so things are easy on the
        # front-end and we don't have to worry about sorting
        'columnIDsMap': {
            column_headers_to_column_ids[column_header]: get_column_header_display(column_header)
            for column_header in original_df.keys()
        },
        'columnFormulasMap': column_formulas,
        'columnFiltersMap': column_filters,
        'columnDtypeMap': column_dtype_map,
        'index': json_obj['index'],
        'dfFormat': df_format,
        'conditionalFormattingResult': get_conditonal_formatting_result(
            state,
            sheet_index,
            original_df,
            df_format['conditional_formats'],
            max_rows=max_rows,
        )

    }


def get_row_data_array(df: pd.DataFrame) -> List[Any]:
    """
    Returns just the data of a dataframe in the 2d array format of [row idx][col idx]
    """
    json_obj = convert_df_to_parsed_json(df)
    return json_obj['data']

def convert_df_to_parsed_json(original_df: pd.DataFrame, max_rows: Optional[int]=MAX_ROWS, max_columns: int=MAX_COLUMNS) -> Dict[str, Any]:
    """
    Returns a dataframe as a json object with the correct formatting
    """
    if max_rows is None:
        df = original_df.copy(deep=True) 
    else:
        # we only show the first max_rows rows!
        df = original_df.head(n=max_rows).copy(deep=True)

    # we only show the first max_columns columns!
    df = df.iloc[: , :max_columns]

    float_columns, date_columns, timedelta_columns = get_float_dt_td_columns(df)
    # We figure out which of the columns contain dates, and we
    # convert them to string columns (for formatting reasons).
    # NOTE: we don't use date_format='iso' in df.to_json call as it appends seconds to the object, 
    # see here: https://stackoverflow.com/questions/52730953/pandas-to-json-output-date-format-in-specific-form
    for column_header in date_columns:
        df[column_header] = df[column_header].dt.strftime('%Y-%m-%d %X')

    # Third, we figure out which of the columns contain timedeltas, and 
    # we format the timedeltas as strings to make them readable
    for column_header in timedelta_columns:
        df[column_header] = df[column_header].apply(lambda x: str(x))

    # Then, we check the index. If it is a datetime or a timedelta, we have to do
    # the same conversions that we did above
    # Then, if we have a datetime index, we update the index to be jsonified better
    if isinstance(df.index, pd.DatetimeIndex):
        df.index = df.index.strftime('%Y-%m-%d %X')
    elif isinstance(df.index, pd.TimedeltaIndex):
        df.index = df.index.to_series().apply(lambda x: str(x))

    json_obj = json.loads(df.to_json(orient="split"))
    # Then, we go through and find all the null values (which are infinities),
    # and set them to 'NaN' for display in the frontend.
    for d in json_obj['data']:
        for idx, e in enumerate(d):
            if e is None:
                d[idx] = 'NaN'

    return json_obj


def get_random_id() -> str:
    """
    Creates a new random ID for the user, which for any given user,
    should only happen once.
    """
    return str(uuid.uuid1())

def get_new_id() -> str:
    return str(uuid.uuid4())

def create_step_id() -> str:
    return '_' + str(randint(10**9, 10**10-1))


def run_command(command_array: List[str]) -> Tuple[str, str]:
    """
    An internal command that should be used to run all commands
    that run on the command line, so that output from failing
    commands can be captured.
    """
    import subprocess
    completed_process = subprocess.run(
        command_array, 
        # NOTE: we do not use the capture_output variable, as this doesn't work before
        # python 3.7
        stdout=subprocess.PIPE, 
        stderr=subprocess.STDOUT,
        # NOTE: we use universal_newlines to get the result back as text, 
        # but we don't use text=True, because we want to work before 3.7 when
        # text was introduced. See here: https://stackoverflow.com/questions/41171791/how-to-suppress-or-capture-the-output-of-subprocess-run
        universal_newlines=True
    )
    # We default the stdout and stderr to empty strings if they are not strings, 
    # to make code that handles them have an easier time (they might be None)
    stdout = completed_process.stdout if isinstance(completed_process.stdout, str) else ''
    stderr = completed_process.stderr if isinstance(completed_process.stderr, str) else ''
    return stdout, stderr

# When you promote a row to a header, the data it can contain might be 
# an numpy type, which json.dumps cannot encode by default. Thus, any
# where we use json.dumps and might have column headers, we need to 
# pass this as a cls=NpEncoder to the json.dumps function
class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.bool_):
            return bool(obj)
        if isinstance(obj, np.integer):
            return int(obj)
        if isinstance(obj, np.floating):
            return float(obj)
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        if isinstance(obj, pd.Timestamp):
            return obj.strftime('%Y-%m-%d %X')
        if isinstance(obj, pd.Timedelta):
            return str(obj)
        return super(NpEncoder, self).default(obj)


def is_prev_version(curr_version: str, benchmark_version: str) -> bool:
    """
    Returns True if the curr_version is previous to the benchmark_version
    Note that this assumes semantic versioning with x.y.z!
    """
    curr_version_parts = curr_version.split('.')
    benchmark_version_parts = benchmark_version.split('.')

    for old_version_part, benchmark_version_part in zip(curr_version_parts, benchmark_version_parts):
        if int(old_version_part) > int(benchmark_version_part):
            # E.g. if we have 0.2.11 and 0.1.11, we want to return early as it's clearly not older!
            return False

        if int(old_version_part) < int(benchmark_version_part):
            return True

    return False

def is_snowflake_connector_python_installed() -> bool:
    try:
        import snowflake.connector
        return True
    except ImportError:
        return False


def is_snowflake_credentials_available() -> bool:
    PYTEST_SNOWFLAKE_USERNAME = os.getenv('PYTEST_SNOWFLAKE_USERNAME')
    PYTEST_SNOWFLAKE_PASSWORD = os.getenv('PYTEST_SNOWFLAKE_PASSWORD')
    PYTEST_SNOWFLAKE_ACCOUNT = os.getenv('PYTEST_SNOWFLAKE_ACCOUNT')

    return PYTEST_SNOWFLAKE_USERNAME is not None and PYTEST_SNOWFLAKE_PASSWORD is not None and PYTEST_SNOWFLAKE_ACCOUNT is not None and \
        PYTEST_SNOWFLAKE_USERNAME != 'None' and PYTEST_SNOWFLAKE_PASSWORD != 'None' and PYTEST_SNOWFLAKE_ACCOUNT != 'None'


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
from typing import Any, Dict, List, Optional, Set, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.empty_code_chunk import EmptyCodeChunk

from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import GraphID


class GraphRenameStepPerformer(StepPerformer):
    """"
    A rename graph step changes the name of a specific graph id
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'graph_rename'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        graph_id = params['graph_id']
        old_graph_tab_name = prev_state.graph_data_dict[graph_id]["graphTabName"]
        params['old_graph_tab_name'] = old_graph_tab_name
        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        graph_id: GraphID = get_param(params, 'graph_id')
        old_graph_tab_name: str = get_param(params, 'old_graph_tab_name')
        new_graph_tab_name: str = get_param(params, 'new_graph_tab_name')

        # Bail early, if there is no change or the new name is empty
        if old_graph_tab_name == new_graph_tab_name or new_graph_tab_name == '' :
            return prev_state, None

        # Create a new step and save the parameters
        post_state = prev_state.copy()

        post_state.graph_data_dict[graph_id]["graphTabName"] = new_graph_tab_name
        
        return post_state, {
            'pandas_processing_time': 0 # No time spent on pandas, only metadata changes
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:

        # Graph steps don't add any generated code to the analysis script. 
        return [
            EmptyCodeChunk(
                prev_state, 
                post_state, 
                'Renamed graph',
                'Renamed a graph',
            )
        ]
    
    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests for the state class
"""
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, DATAFRAME_SOURCE_PASSED, State
import pandas as pd

def test_state_can_add_df_to_end():
    df = pd.DataFrame({'A': [123]})
    state = State([df])
    state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED)
    for key, value in state.__dict__.items():
        if isinstance(value, list):
            assert len(value) == 2
    
    assert state.df_sources == [DATAFRAME_SOURCE_PASSED, DATAFRAME_SOURCE_IMPORTED]

def test_state_can_add_df_to_middle():
    df = pd.DataFrame({'A': [123]})
    state = State([df])
    state.add_df_to_state(df, DATAFRAME_SOURCE_IMPORTED, sheet_index=0)
    for key, value in state.__dict__.items():
        if isinstance(value, list):
            assert len(value) == 1
    
    assert state.df_sources == [DATAFRAME_SOURCE_IMPORTED]




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
This file contains shared objects for when the column renames
were moved from the codebase, and so we needed to add additional
steps to make the analysis backwards compatible.
"""

INITIAL_BULK_OLD_RENAME_STEP = {
    "step_version": 1, 
    "step_type": "bulk_old_rename", 
    "params": {
        "move_to_deprecated_id_algorithm": True
    }
}

BULK_OLD_RENAME_STEP = {
    "step_version": 1, 
    "step_type": "bulk_old_rename", 
    "params": {}
}


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import DeleteColumnsCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import RenameColumnsCodeChunk
from mitosheet.parser import parse_formula
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import \
    column_header_to_transpiled_code
from mitosheet.types import ColumnHeader, ColumnID, FormulaAppliedToType


class AddColumnSetFormulaCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_id: ColumnID, formula_label: Union[str, bool, int, float], index_labels_formula_is_applied_to: FormulaAppliedToType, column_header: ColumnHeader, column_header_index: int, new_formula: str):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_id = column_id
        self.column_header = column_header
        self.formula_label = formula_label
        self.index_labels_formula_is_applied_to = index_labels_formula_is_applied_to
        self.column_header_index = column_header_index
        self.new_formula = new_formula

        self.df_name = self.post_state.df_names[self.sheet_index]

    def get_display_name(self) -> str:
        return 'Added column'
    
    def get_description_comment(self) -> str:
        return f'Added column {column_header_to_transpiled_code(self.column_header)}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        python_code, _, _, _ = parse_formula(
            self.new_formula, 
            self.column_header,
            self.formula_label,
            self.index_labels_formula_is_applied_to,
            self.post_state.dfs[self.sheet_index],
            df_name=self.post_state.df_names[self.sheet_index],
            include_df_set=False
        )

        transpiled_column_header = column_header_to_transpiled_code(self.column_header)
        column_header_index = self.column_header_index
        return [
            f'{self.df_name}.insert({column_header_index}, {transpiled_column_header}, {python_code})'
        ], []

    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_columns_code_chunk(self, other_code_chunk: DeleteColumnsCodeChunk) -> Optional["CodeChunk"]:
        # Make sure the sheet index matches up first
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None
        
        # Check to see if the column ids overlap
        added_column_id = self.post_state.column_ids.get_column_id_by_header(self.sheet_index, self.column_header)
        deleted_column_ids = other_code_chunk.column_ids

        if added_column_id in deleted_column_ids and len(deleted_column_ids) == 1:
            return NoOpCodeChunk(
                self.prev_state, 
                other_code_chunk.post_state, 
            )
        elif added_column_id in deleted_column_ids:
            new_deleted_column_ids = copy(deleted_column_ids)
            new_deleted_column_ids.remove(added_column_id)

            return DeleteColumnsCodeChunk(
                self.prev_state,
                other_code_chunk.post_state,
                self.sheet_index,
                new_deleted_column_ids
            )
        
        return None

    def _combine_right_with_rename_columns_code_chunk(self, other_code_chunk: RenameColumnsCodeChunk) -> Optional["CodeChunk"]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        added_column_id = self.column_id
        column_ids_to_new_column_headers = other_code_chunk.column_ids_to_new_column_headers

        if added_column_id in column_ids_to_new_column_headers and len(column_ids_to_new_column_headers) == 1:
            return AddColumnSetFormulaCodeChunk(
                self.prev_state,
                other_code_chunk.post_state,
                self.sheet_index,
                self.column_id,
                self.formula_label,
                self.index_labels_formula_is_applied_to,
                column_ids_to_new_column_headers[added_column_id],
                self.column_header_index,
                self.new_formula
            )

        return None
        
    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_columns_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, RenameColumnsCodeChunk):
            return self._combine_right_with_rename_columns_code_chunk(other_code_chunk)

        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

import os
from typing import Any, Dict, List, Optional, Tuple
import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import column_header_to_transpiled_code
from mitosheet.utils import is_prev_version

# Note: These defaults must be the same as the pandas.read_csv defaults
DEFAULT_ENCODING = 'utf-8'
DEFAULT_DELIMETER = ','
DEFAULT_DECIMAL = '.'
DEFAULT_SKIPROWS = 0
DEFAULT_ERROR_BAD_LINES = True

def get_read_csv_params(delimeter: str, encoding: str, decimal: Optional[str], skiprows: Optional[int], error_bad_lines: Optional[bool]) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    if encoding != DEFAULT_ENCODING:
        params['encoding'] = encoding
    if delimeter != DEFAULT_DELIMETER:
        params['sep'] = delimeter
    if decimal is not None and decimal != DEFAULT_DECIMAL: 
        params['decimal'] = decimal
    if skiprows is not None and skiprows != DEFAULT_SKIPROWS:
        params['skiprows'] = skiprows
    if error_bad_lines is not None and error_bad_lines != DEFAULT_ERROR_BAD_LINES:
        # See here: https://datascientyst.com/drop-bad-lines-with-read_csv-pandas/
        if is_prev_version(pd.__version__, '1.3.0'):
            params['error_bad_lines'] = error_bad_lines
        else:
            params['on_bad_lines'] = 'skip'
    
    return params


def generate_read_csv_code(file_name: str, df_name: str, delimeter: str, encoding: str, decimal: Optional[str], skiprows: Optional[int], error_bad_lines: Optional[bool]) -> str:
    """
    Helper function for generating minimal read_csv code 
    depending on the delimeter and the encoding of a file
    """

    params = get_read_csv_params(delimeter, encoding, decimal=decimal, skiprows=skiprows, error_bad_lines=error_bad_lines)
    params_string = ', '.join(f'{key}={column_header_to_transpiled_code(value)}' for key, value in params.items())

    return f'{df_name} = pd.read_csv(r\'{file_name}\'{", " if len(params_string) > 0 else ""}{params_string})'


class SimpleImportCodeChunk(CodeChunk):

    
    def __init__(self, prev_state: State, post_state: State, file_names: List[str], file_delimeters: List[str], file_encodings: List[str], file_decimals: List[str], file_skiprows: List[int], file_error_bad_lines: List[bool]):
        super().__init__(prev_state, post_state)
        self.file_names = file_names
        self.file_delimeters = file_delimeters
        self.file_encodings = file_encodings
        self.file_decimals = file_decimals
        self.file_skiprows = file_skiprows
        self.file_error_bad_lines = file_error_bad_lines

    def get_display_name(self) -> str:
        return 'Imported'
    
    def get_description_comment(self) -> str:
        base_names = [os.path.basename(path) for path in self.file_names]
        return f'Imported {", ".join(base_names)}'

    def get_code(self) -> Tuple[List[str], List[str]]:

        code = []

        index = 0
        for file_name, df_name in zip(self.file_names, self.post_state.df_names[len(self.post_state.df_names) - len(self.file_names):]):

            delimeter = self.file_delimeters[index]
            encoding = self.file_encodings[index]
            decimal = self.file_decimals[index]
            skiprows = self.file_skiprows[index]
            error_bad_lines = self.file_error_bad_lines[index]

            code.append(
                generate_read_csv_code(file_name, df_name, delimeter, encoding, decimal, skiprows, error_bad_lines)
            )
            
            index += 1

        return code, ['import pandas as pd']

    def get_created_sheet_indexes(self) -> List[int]:
        return [i for i in range(len(self.post_state.dfs) - len(self.file_names), len(self.post_state.dfs))]

    def _combine_right_simple_import(self, other_code_chunk: "SimpleImportCodeChunk") -> Optional["CodeChunk"]:
        # We can easily combine simple imports, so we do so
        file_names = self.file_names + other_code_chunk.file_names
        new_file_delimeters = self.file_delimeters + other_code_chunk.file_delimeters
        new_file_encodings = self.file_encodings + other_code_chunk.file_encodings
        new_file_decimals = self.file_decimals + other_code_chunk.file_decimals
        new_file_skiprows = self.file_skiprows + other_code_chunk.file_skiprows
        new_error_bad_lines = self.file_error_bad_lines + other_code_chunk.file_error_bad_lines

        return SimpleImportCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            file_names,
            new_file_delimeters,
            new_file_encodings,
            new_file_decimals,
            new_file_skiprows,
            new_error_bad_lines
        )

    def combine_right(self, other_code_chunk: "CodeChunk") -> Optional["CodeChunk"]:
        if isinstance(other_code_chunk, SimpleImportCodeChunk):
            return self._combine_right_simple_import(other_code_chunk)

        return None



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from copy import copy
import json
import os
from typing import Any, Dict, List

import pandas as pd
from mitosheet.types import StepsManagerType
from mitosheet.step_performers.import_steps import is_import_step_type
from mitosheet.saved_analyses import read_and_upgrade_analysis
from mitosheet.api.get_imported_files_and_dataframes_from_current_steps import get_import_data_with_single_import_list


def get_step_import_data_from_saved_analysis(analysis_name: str) -> List[Dict[str, Any]]:
    # Read in the data, and turn it into the correct format
    saved_analysis = read_and_upgrade_analysis(analysis_name)
    if saved_analysis is None:
        return []
    else:
        # Find all the import steps, and turn them into imports
        step_import_data_list = []
        for step_data in saved_analysis['steps_data']:
            step_type = step_data['step_type']
            params = step_data['params']
            if is_import_step_type(step_type):
                step_import_data_list.append({
                    'step_id': 'fake_id', 
                    'imports': get_import_data_with_single_import_list(step_type, params)
                })
        
        return step_import_data_list


def get_imported_files_and_dataframes_from_analysis_name(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Returns a list all the imported files and dataframes, and their import params. Does so by turning
    the import into a list of imports that only import a single dataframe, as this is convenient
    for what we work with on the frontend.

    A mirror of get_imported_files_and_dataframes_from_current_steps, but works from saved analysis data, and is used
    when a replayed analysis is failed - as there are no steps in the steps_manager currently.

    In the case where we're getting import data from a saved analysis, we do not have step ids to use.
    But we don't need step ids, as when we replay an analysis we can just replace import events based
    on the index of the import. As such, don't worry about step_ids in the UpdateImportPreReplay
    case, as this uses the replay_analysis function and matched based on indexes.
    """
    analysis_name = params["analysis_name"]

    # We turn import steps into a version of the step that just creates a single dataframe
    # as this is what is easiest to work with on the frontend
    return json.dumps(get_step_import_data_from_saved_analysis(analysis_name))



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List

from mitosheet.saved_analyses.step_upgraders.utils_column_header_to_column_id import \
    replace_headers_with_id


def upgrade_set_dataframe_format_1_to_2(step: Dict[str, Any], later_steps: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Adds the conditional format part of the params.

    OLD: {
        "step_version": 1, 
        "step_type": "set_dataframe_format", 
        "params": {
            "sheet_index": 0, 
            "df_format": {
                "columns": {}, 
                "headers": {"color": "#FFFFFF", "backgroundColor": "#549D3A"}, 
                "rows": {"even": {"color": "#494650", "backgroundColor": "#D0E3C9"}, "odd": {"color": "#494650"}}, 
                "border": {}}
            }
        }
    }

    NEW: {
        "step_version": 2, 
        "step_type": "set_dataframe_format", 
        "params": {
            "sheet_index": 0, 
            "df_format": {
                "columns": {}, 
                "headers": {"color": "#FFFFFF", "backgroundColor": "#549D3A"}, 
                "rows": {"even": {"color": "#494650", "backgroundColor": "#D0E3C9"}, "odd": {"color": "#494650"}}, 
                "border": {}}
                "conditional_formats": []
            }
        }
    }
    """
    params = step['params']
    params['df_format']['conditional_formats'] = []

    return [{
        "step_version": 2, 
        "step_type": "set_dataframe_format", 
        "params": params
    }] + later_steps




#!/usr/bin/env python
# coding: utf-8

from copy import copy
# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Dict, List, Optional, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.state import State
from mitosheet.transpiler.transpile_utils import param_dict_to_code


class SnowflakeImportCodeChunk(CodeChunk):

    def __init__(
        self, 
        prev_state: State, 
        post_state: State, 
        connection_params_dict: Dict[str, str], 
        sql_queries: List[str],
    ):
        super().__init__(prev_state, post_state)
        self.connection_params_dict = connection_params_dict
        self.sql_queries = sql_queries

    def get_display_name(self) -> str:
        return 'Import from Snowflake'
    
    def get_description_comment(self) -> str:
        return "Imported dataframe from Snowflake"

    def get_code(self) -> Tuple[List[str], List[str]]:

        connection_param_transpiled_code = param_dict_to_code(self.connection_params_dict)

        snowflake_connection_code = [
            f'con = snowflake.connector.connect({connection_param_transpiled_code})',
            '',
            'cur = con.cursor()',
        ]

        df_names = self.post_state.df_names[len(self.prev_state.df_names):]

        df_creation_code = ['']
        for sql_query, df_name in zip(self.sql_queries, df_names):
            df_creation_code.append(f'cur.execute(\'{sql_query}\')')
            df_creation_code.append(f'{df_name} = cur.fetch_pandas_all()')
            df_creation_code.append('')

        close_connection_code = ['con.close()']

        all_code = snowflake_connection_code + df_creation_code + close_connection_code
                        
        return all_code, ['import snowflake.connector']

    def get_created_sheet_indexes(self) -> List[int]:
        return [i for i in range(len(self.post_state.dfs) - len(self.sql_queries), len(self.post_state.dfs))]


    def _combine_right_with_snowflake_import_code_chunk(self, other_code_chunk: "SnowflakeImportCodeChunk") -> Optional["SnowflakeImportCodeChunk"]:
        if not self.params_match(other_code_chunk, ['connection_params_dict']):
            return None
        
        all_sql_queries = copy(self.sql_queries) # Make sure to copy this so we don't get weird bugs w/ duplication
        all_sql_queries.extend(other_code_chunk.sql_queries)
        
        return SnowflakeImportCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            self.connection_params_dict,
            all_sql_queries,
        )        

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, SnowflakeImportCodeChunk):
            return self._combine_right_with_snowflake_import_code_chunk(other_code_chunk)
            
        return None
        


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from copy import copy
from typing import Any, Dict, List, Optional, Tuple

from mitosheet.code_chunks.add_column_set_formula_code_chunk import \
    AddColumnSetFormulaCodeChunk
from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.no_op_code_chunk import NoOpCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.delete_column_code_chunk import \
    DeleteColumnsCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.rename_columns_code_chunk import \
    RenameColumnsCodeChunk
from mitosheet.code_chunks.step_performers.column_steps.set_column_formula_code_chunk import \
    SetColumnFormulaCodeChunk
from mitosheet.state import State
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE
from mitosheet.transpiler.transpile_utils import \
    column_header_to_transpiled_code


class AddColumnCodeChunk(CodeChunk):

    def __init__(self, prev_state: State, post_state: State, sheet_index: int, column_header: str, column_header_index: int):
        super().__init__(prev_state, post_state)
        self.sheet_index = sheet_index
        self.column_header = column_header
        self.column_header_index = column_header_index

        self.column_id = self.post_state.column_ids.get_column_id_by_header(self.sheet_index, self.column_header)
        self.df_name: str = self.post_state.df_names[self.sheet_index]


    def get_display_name(self) -> str:
        return 'Added column'
    
    def get_description_comment(self) -> str:
        return f'Added column {self.column_header}'

    def get_code(self) -> Tuple[List[str], List[str]]:
        transpiled_column_header = column_header_to_transpiled_code(self.column_header)
        return [
            f'{self.df_name}.insert({self.column_header_index}, {transpiled_column_header}, 0)'
        ], []

    
    def get_edited_sheet_indexes(self) -> List[int]:
        return [self.sheet_index]

    def _combine_right_with_delete_column_code_chunk(self, other_code_chunk: DeleteColumnsCodeChunk) -> Optional[CodeChunk]:
        # Make sure the sheet index matches up first
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None
        
        # Check to see if the column ids overlap
        sheet_index = self.sheet_index
        added_column_header = self.column_header
        added_column_id = self.post_state.column_ids.get_column_id_by_header(sheet_index, added_column_header)
        deleted_column_ids = other_code_chunk.column_ids

        if added_column_id in deleted_column_ids and len(deleted_column_ids) == 1:
            return NoOpCodeChunk(
                self.prev_state, 
                other_code_chunk.post_state
            )
        elif added_column_id in deleted_column_ids:
            new_deleted_column_ids = copy(deleted_column_ids)
            new_deleted_column_ids.remove(added_column_id)

            return DeleteColumnsCodeChunk(
                self.prev_state,
                other_code_chunk.post_state,
                sheet_index,
                new_deleted_column_ids
            )
        
        return None
    
    def _combine_right_with_rename_columns_code_chunk(self, other_code_chunk: RenameColumnsCodeChunk) -> Optional[CodeChunk]:
        # Make sure the sheet index matches up first
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None
        
        # Check to see if the column ids overlap
        added_column_id = self.post_state.column_ids.get_column_id_by_header(self.sheet_index, self.column_header)
        column_ids_to_new_column_headers = other_code_chunk.column_ids_to_new_column_headers

        if added_column_id in column_ids_to_new_column_headers and len(column_ids_to_new_column_headers) == 1:
            # Handle the case where the is just one column added here
            new_column_header = column_ids_to_new_column_headers[added_column_id]
            return AddColumnCodeChunk(
                self.prev_state,
                other_code_chunk.post_state,
                self.sheet_index,
                new_column_header,
                self.column_header_index 
            )
        else:
            # We'd prefer to keep all the renames together in this case, although
            # this will not happen often
            return None
    
    def _combine_right_with_set_column_formula_code_chunk(self, other_code_chunk: SetColumnFormulaCodeChunk) -> Optional[CodeChunk]:
        if not self.params_match(other_code_chunk, ['sheet_index']):
            return None

        # If this does not set the entire column, we can't combine -- the insert doesn't work
        if other_code_chunk.index_labels_formula_is_applied_to['type'] != FORMULA_ENTIRE_COLUMN_TYPE:
            return None

        sheet_index = self.sheet_index
        added_column_header = self.column_header
        added_column_id = self.post_state.column_ids.get_column_id_by_header(sheet_index, added_column_header)
        
        if added_column_id != other_code_chunk.column_id:
            return None

        return AddColumnSetFormulaCodeChunk(
            self.prev_state,
            other_code_chunk.post_state,
            self.sheet_index,
            added_column_id,
            other_code_chunk.formula_label,
            other_code_chunk.index_labels_formula_is_applied_to,
            self.column_header,
            self.column_header_index,
            other_code_chunk.new_formula
        )

    def combine_right(self, other_code_chunk: CodeChunk) -> Optional[CodeChunk]:
        if isinstance(other_code_chunk, DeleteColumnsCodeChunk):
            return self._combine_right_with_delete_column_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, RenameColumnsCodeChunk):
            return self._combine_right_with_rename_columns_code_chunk(other_code_chunk)
        elif isinstance(other_code_chunk, SetColumnFormulaCodeChunk):
            return self._combine_right_with_set_column_formula_code_chunk(other_code_chunk)

        return None


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
import json
import os
from typing import Any, Dict

import pandas as pd
from mitosheet.types import StepsManagerType


def get_excel_file_metadata(params: Dict[str, Any], steps_manager: StepsManagerType) -> str:
    """
    Given a 'file_name' that should be an XLSX file, 
    will get the metadata for that XLSX file. 

    For now, this is just the sheets this file contains, 
    but in the future we may be able to request more about 
    the workbook
    """
    file_path = params['file_path']

    file = pd.ExcelFile(file_path, engine='openpyxl')
    sheet_names = file.sheet_names

    return json.dumps({
        'sheet_names': sheet_names,
        'size': os.path.getsize(file_path)
    })





import subprocess
import os

import pandas as pd
import pytest

from mitosheet.mito_backend import get_mito_frontend_code
from mitosheet.steps_manager import StepsManager
from mitosheet.tests.test_utils import create_mito_wrapper_dfs


# See here: https://www.tutorialspoint.com/json_simple/json_simple_escape_characters.htm
STRINGS_TO_TEST = [
    ('	'),
    ('Normal'),
    ('space in the text'),
    ('学	医'),
    ('this is \\text'),
    ('\n'),
    ('\t'),
    ('\r'),
    ('`'),
    ("'"),
    ('"'),
    ('&'),
    ('\b'),
    ('\f'),
]


def write_test_code(file: str, string: str) -> None:
    df = pd.DataFrame({'A': [string]})
    mito = create_mito_wrapper_dfs(df)
    code = get_mito_frontend_code('a', 'a', 'a', mito.mito_backend)
    with open(file, 'w+') as f:
        f.write(code)

# Skip if not not defined
@pytest.mark.parametrize('string', STRINGS_TO_TEST)
def test_mito_frontend_is_valid_code(tmp_path, string):
    
    file = tmp_path / 'out.js'
    write_test_code(file, string)
    p = subprocess.Popen(['node', file,], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    (_, err) = p.communicate()
    p.wait()

    # we want to make sure that there are no failures in parsing, that it runs up to the 
    # ReferenceError: document is not defined 
    assert 'SyntaxError' not in err.decode('utf-8') 
    assert 'ReferenceError' in err.decode('utf-8') 


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

# We export any function that is useful throughout the codebase

from mitosheet.user.create import initialize_user
from mitosheet.user.db import (
    get_user_field, set_user_field 
)
from mitosheet.user.utils import (
    is_running_test, is_on_kuberentes_mito, is_local_deployment,
    should_upgrade_mitosheet, is_pro
)
from mitosheet.user.schemas import (
    UJ_INTENDED_BEHAVIOR, UJ_CLOSED_FEEDBACK, UJ_MITOSHEET_LAST_FIVE_USAGES,
    UJ_USER_JSON_VERSION, UJ_STATIC_USER_ID, UJ_USER_SALT, UJ_USER_EMAIL, 
    UJ_RECEIVED_TOURS, UJ_FEEDBACKS, UJ_FEEDBACKS_V2, UJ_MITOSHEET_CURRENT_VERSION, 
    UJ_MITOSHEET_LAST_UPGRADED_DATE, UJ_MITOSHEET_LAST_FIFTY_USAGES, UJ_MITOSHEET_PRO,
    UJ_EXPERIMENT
)




#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.import_steps.excel_range_import_code_chunk import \
    ExcelRangeImportCodeChunk
from mitosheet.errors import make_upper_left_corner_value_not_found_error
from mitosheet.excel_utils import get_column_from_column_index, get_col_and_row_indexes_from_range, get_table_range_from_upper_left_corner_value
from mitosheet.state import DATAFRAME_SOURCE_IMPORTED, State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ExcelRangeImport
from mitosheet.utils import get_valid_dataframe_name

EXCEL_RANGE_IMPORT_TYPE_RANGE = 'range'
EXCEL_RANGE_IMPORT_TYPE_UPPER_LEFT = 'upper left corner value'


class ExcelRangeImportStepPerformer(StepPerformer):
    """
    Allows you to import multiple ranges from a single excel sheet.
    """

    @classmethod
    def step_version(cls) -> int:
        return 1

    @classmethod
    def step_type(cls) -> str:
        return 'excel_range_import'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        file_path: str = get_param(params, 'file_path')
        sheet_name: str = get_param(params, 'sheet_name')
        range_imports: List[ExcelRangeImport] = get_param(params, 'range_imports')
        
        post_state = prev_state.copy() 

        pandas_start_time = perf_counter()

        sheet_index_to_df_range: Dict[int, str] = {}
        for range_import in range_imports:
            _range: Optional[str]
            if range_import['type'] == EXCEL_RANGE_IMPORT_TYPE_RANGE:
                _range = range_import['value']
            else:
                _range = get_table_range_from_upper_left_corner_value(file_path, sheet_name, range_import['value'])

            if _range is None:
                raise make_upper_left_corner_value_not_found_error(range_import['value'], False)

            ((start_col_index, start_row_index), (end_col_index, end_row_index)) = get_col_and_row_indexes_from_range(_range)
            nrows = end_row_index - start_row_index
            usecols = get_column_from_column_index(start_col_index) + ':' + get_column_from_column_index(end_col_index)

            df = pd.read_excel(file_path, sheet_name=sheet_name, skiprows=start_row_index, nrows=nrows, usecols=usecols)
            final_df_name = get_valid_dataframe_name(post_state.df_names, range_import['df_name'])
            post_state.add_df_to_state(
                df,
                DATAFRAME_SOURCE_IMPORTED,
                df_name=final_df_name
            )

            sheet_index_to_df_range[len(post_state.dfs) - 1] = _range

        pandas_processing_time = perf_counter() - pandas_start_time

        return post_state, {
            'pandas_processing_time': pandas_processing_time,
            'new_sheet_index_to_df_range': sheet_index_to_df_range,
            'result': {
                # TODO: fill in the result, when we make the frontend
            }
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            ExcelRangeImportCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'file_path'),
                get_param(params, 'sheet_name'),
                (execution_data if execution_data is not None else dict()).get('new_sheet_index_to_df_range', {})
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {-1}
    


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.

from time import perf_counter
from typing import Any, Dict, List, Optional, Set, Tuple

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.sort_code_chunk import SortCodeChunk
from mitosheet.errors import (make_invalid_sort_error,
                              raise_error_if_column_ids_do_not_exist)
from mitosheet.state import State
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.types import ColumnID

# CONSTANTS USED IN THE SORT STEP ITSELF
SORT_DIRECTION_ASCENDING = 'ascending'
SORT_DIRECTION_DESCENDING = 'descending'
SORT_DIRECTION_NONE = 'none'

class SortStepPerformer(StepPerformer):
    """
    Allows you to sort a df based on key column, in either
    ascending or descending order.
    """

    @classmethod
    def step_version(cls) -> int:
        return 2

    @classmethod
    def step_type(cls) -> str:
        return 'sort'

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        column_id: ColumnID = get_param(params, 'column_id')
        sort_direction: str = get_param(params, 'sort_direction')

        raise_error_if_column_ids_do_not_exist(
            'sort',
            prev_state,
            sheet_index,
            column_id
        )

        column_header = prev_state.column_ids.get_column_header_by_id(sheet_index, column_id)

        # We make a new state to modify it
        post_state = prev_state.copy(deep_sheet_indexes=[sheet_index])

        try: 
            pandas_start_time = perf_counter()
            if sort_direction != SORT_DIRECTION_NONE:
                new_df = prev_state.dfs[sheet_index].sort_values(by=column_header, ascending=(sort_direction == SORT_DIRECTION_ASCENDING), na_position=('first' if sort_direction == SORT_DIRECTION_ASCENDING else 'last'))
            else:
                # We notably let the user sort by the "none" direction, which effectively allows the user
                # to unapply the sort by toggling the sort button after they click it once
                new_df = prev_state.dfs[sheet_index]
            
            pandas_processing_time = perf_counter() - pandas_start_time
            post_state.dfs[sheet_index] = new_df
        except TypeError as e:
            # A NameError occurs when you try to sort a column with incomparable 
            # dtypes (ie: a column with strings and floats)
            print(e)
            # Generate an error informing the user
            raise make_invalid_sort_error(column_header)

        return post_state, {
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            SortCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'column_id'),
                get_param(params, 'sort_direction')
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        return {get_param(params, 'sheet_index')}



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from mitosheet.startup.startup_utils import IMPORT_MITOSHEET_FILE_CONTENTS as mitosheet_file_contents
import os
import subprocess

IMPORT_MITOSHEET_FILE_PATH = os.path.join(os.path.expanduser("~"), '.ipython', 'profile_default', 'startup', 'import_mitosheet.py')

def test_create_startup_file():
    subprocess.run(['python', '-m', 'mitosheet', 'turnondataframebutton'])
    assert os.path.exists(IMPORT_MITOSHEET_FILE_PATH) 

    subprocess.run(['python', '-m', 'mitosheet', 'turnoffdataframebutton'])
    assert not os.path.exists(IMPORT_MITOSHEET_FILE_PATH) 

    subprocess.run(['python', '-m', 'mitosheet', 'turnoffdataframebutton'])
    subprocess.run(['python', '-m', 'mitosheet', 'turnondataframebutton'])
    assert os.path.exists(IMPORT_MITOSHEET_FILE_PATH) 


def test_file_contents_in_sync():
    import sys
    sys.path.insert(0, '../mitoinstaller')
    from mitoinstaller.create_startup_file import IMPORT_MITOSHEET_FILE_CONTENTS as installer_file_contents
    assert installer_file_contents == mitosheet_file_contents


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains tests for edit events.
"""
import random
from typing import Type

from mitosheet.step_performers import (
    STEP_PERFORMERS,
    StepPerformer,
    PivotStepPerformer,
    FilterStepPerformer,
    SortStepPerformer,
    ReorderColumnStepPerformer,
    AddColumnStepPerformer,
    SetColumnFormulaStepPerformer,
    SetCellValueStepPerformer,
    MergeStepPerformer,
    DeleteColumnStepPerformer,
    RenameColumnStepPerformer,
    ChangeColumnDtypeStepPerformer,
    SimpleImportStepPerformer,
    DataframeDeleteStepPerformer,
    DataframeDuplicateStepPerformer,
    DataframeRenameStepPerformer,
    BulkOldRenameStepPerformer, 
    ExcelImportStepPerformer,
    DropDuplicatesStepPerformer,
    GraphStepPerformer,
    ConcatStepPerformer,
    ExcelRangeImportStepPerformer,
    ExportToFileStepPerformer,
    ResetIndexStepPerformer,
    SnowflakeImportStepPerformer
)
from mitosheet.step_performers.import_steps.dataframe_import import DataframeImportStepPerformer
from mitosheet.step_performers.delete_row import DeleteRowStepPerformer
from mitosheet.step_performers.column_steps.split_text_to_columns import SplitTextToColumnsStepPerformer
from mitosheet.step_performers.fill_na import FillNaStepPerformer
from mitosheet.step_performers.graph_steps.graph_delete import GraphDeleteStepPerformer
from mitosheet.step_performers.graph_steps.graph_duplicate import GraphDuplicateStepPerformer
from mitosheet.step_performers.graph_steps.graph_rename import GraphRenameStepPerformer
from mitosheet.step_performers.melt import MeltStepPerformer
from mitosheet.enterprise.step_performers.one_hot_encoding import OneHotEncodingStepPerformer
from mitosheet.step_performers.promote_row_to_header import PromoteRowToHeaderStepPerformer
from mitosheet.pro.step_performers.set_dataframe_format import SetDataframeFormatStepPerformer
from mitosheet.step_performers.transpose import TransposeStepPerformer

def check_step(
        step_performer: Type[StepPerformer], 
        step_version: int, 
        step_type: str
    ) -> None:
    """
    Helper function that checks a given step definition against the 
    expected step_version, step_type, and params for that step. 

    Throws an assertion error if any of them do not match! 
    """
    assert step_performer.step_version() == step_version
    assert step_performer.step_type() == step_type

def test_params_static():
    """
    NOTE: This is a regression test! Before changing this test to make it pass, talk to 
    the engineering team and make sure you know what you're doing. 

    Remeber:
    1. Each Mito analysis is written to a file, where each step is saved with its
       parameters.
    2. If the _name_ of the step, or the _parameters to the step_ change, then this
       will break existing user analyses. 
    
    Thus, this test is to make sure that we _know_ when we're breaking things. 
    
    However, note that there are ways to break existing user analyses other than this. 
    For example, if you change how the group step flattens headers, this may cause
    issues if the user then later renames one of those flattened columns. So, this
    regression test is not perfect, but it is a good start!
    """

    check_step(
        AddColumnStepPerformer,
        2,
        'add_column'
    )

    check_step(
        DeleteColumnStepPerformer,
        3,
        'delete_column'
    )

    check_step(
        RenameColumnStepPerformer,
        2,
        'rename_column'
    )

    check_step(
        FilterStepPerformer,
        4,
        'filter_column'
    )
    
    check_step(
        PivotStepPerformer,
        8,
        'pivot'
    )

    check_step(
        MergeStepPerformer,
        4,
        'merge'
    )

    check_step(
        ReorderColumnStepPerformer,
        2,
        'reorder_column'
    )

    check_step(
        SetColumnFormulaStepPerformer,
        4,
        'set_column_formula'
    )

    check_step(
        SortStepPerformer,
        2,
        'sort'
    )


    check_step(
        ChangeColumnDtypeStepPerformer,
        3,
        'change_column_dtype'
    )

    check_step(
        SetDataframeFormatStepPerformer,
        2,
        'set_dataframe_format'
    )

    check_step(
        SimpleImportStepPerformer,
        2,
        'simple_import'
    )

    check_step(
        DataframeDeleteStepPerformer,
        1,
        'dataframe_delete'
    )

    check_step(
        DataframeDuplicateStepPerformer,
        1,
        'dataframe_duplicate'
    )

    check_step(
        DataframeRenameStepPerformer,
        1,
        'dataframe_rename'
    )

    check_step(
        SetCellValueStepPerformer,
        1,
        'set_cell_value'
    )

    check_step(
        BulkOldRenameStepPerformer,
        1,
        'bulk_old_rename'
    )

    check_step(
        ExcelImportStepPerformer,
        1,
        'excel_import'
    )

    check_step(
        DropDuplicatesStepPerformer,
        1,
        'drop_duplicates'
    )

    check_step(
        GraphStepPerformer,
        4,
        'graph'
    )

    check_step(
        GraphDeleteStepPerformer,
        1,
        'graph_delete'
    )

    check_step(
        GraphDuplicateStepPerformer,
        1,
        'graph_duplicate'
    )

    check_step(
        GraphRenameStepPerformer,
        1,
        'graph_rename'
    )

    check_step(
        ConcatStepPerformer,
        1,
        'concat'
    )

    check_step(
        FillNaStepPerformer,
        1,
        'fill_na'
    )

    check_step(
        DeleteRowStepPerformer,
        1,
        'delete_row'
    )

    check_step(
        PromoteRowToHeaderStepPerformer,
        1,
        'promote_row_to_header'
    )

    check_step(
        SplitTextToColumnsStepPerformer,
        1,
        'split_text_to_columns'
    )

    check_step(
        TransposeStepPerformer,
        1,
        'transpose'
    )

    check_step(
        MeltStepPerformer,
        1,
        'melt'
    )

    check_step(
        OneHotEncodingStepPerformer,
        1,
        'one_hot_encoding'
    )

    check_step(
        DataframeImportStepPerformer,
        1,
        'dataframe_import'
    )

    check_step(
        ResetIndexStepPerformer,
        1,
        'reset_index'
    )

    check_step(
        ExcelRangeImportStepPerformer,
        1,
        'excel_range_import'
    )

    check_step(
        ExportToFileStepPerformer,
        1,
        'export_to_file'
    )

    check_step(
        SnowflakeImportStepPerformer,
        1,
        'snowflake_import'
    )

    assert len(STEP_PERFORMERS) == 36


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
"""
Contains functions for upgrading analyses!

NOTE: when adding a specific function for upgrading one version of a step to the other, the
format of the function should be:

def upgrade_<old_step_type>_<old_step_version>_to_<new_step_type>_<new_step_version>(step):
    ....
"""

from copy import copy
from typing import Any, Callable, Dict, List, Optional

from mitosheet._version import __version__, package_name
from mitosheet.saved_analyses.schema_utils import (upgrade_saved_analysis_format_to_steps_data)
from mitosheet.saved_analyses.step_upgraders.add_column import \
    upgrade_add_column_1_to_add_column_2
from mitosheet.saved_analyses.step_upgraders.change_column_dtype import (
    upgrade_change_column_dtype_1_to_2, upgrade_change_column_dtype_2_to_3)
from mitosheet.saved_analyses.step_upgraders.change_column_format import \
    upgrade_change_column_format_1_to_remove
from mitosheet.saved_analyses.step_upgraders.delete_column import (
    upgrade_delete_column_1_to_2, upgrade_delete_column_2_to_3)
from mitosheet.saved_analyses.step_upgraders.filter import (
    update_filter_column_1_to_filter_column_2, upgrade_filter_column_2_to_3,
    upgrade_filter_column_3_to_4)
from mitosheet.saved_analyses.step_upgraders.graph import (
    upgrade_graph_1_to_2, upgrade_graph_2_to_3, upgrade_graph_3_to_4)
from mitosheet.saved_analyses.step_upgraders.merge import (
    upgrade_merge_1_to_merge_2, upgrade_merge_2_to_3, upgrade_merge_3_to_4)
from mitosheet.saved_analyses.step_upgraders.pivot import (
    upgrade_group_1_to_pivot_2, upgrade_pivot_2_to_pivot_3,
    upgrade_pivot_3_to_4, upgrade_pivot_4_to_5_and_rename,
    upgrade_pivot_5_to_6, upgrade_pivot_6_to_7, upgrade_pivot_7_to_8)
from mitosheet.saved_analyses.step_upgraders.rename_column import \
    upgrade_rename_column_1_to_2
from mitosheet.saved_analyses.step_upgraders.reorder_column import \
    upgrade_reorder_column_1_to_2
from mitosheet.saved_analyses.step_upgraders.set_column_formula import \
    upgrade_set_column_formula_1_to_2, upgrade_set_column_formula_2_to_3, upgrade_set_column_formula_3_to_4
from mitosheet.saved_analyses.step_upgraders.set_dataframe_format import \
    upgrade_set_dataframe_format_1_to_2
from mitosheet.saved_analyses.step_upgraders.simple_import import \
    upgrade_simple_import_1_to_2_and_rename
from mitosheet.saved_analyses.step_upgraders.sort import upgrade_sort_1_to_2
from mitosheet.saved_analyses.step_upgraders.utils_rename_column_headers import \
    INITIAL_BULK_OLD_RENAME_STEP
from mitosheet.utils import is_prev_version

"""
STEP_UPGRADES_FUNCTION_MAPPING mapping contains a mapping of all steps that need to be upgraded. A step
x at version y needs to be upgraded if STEP_UPGRADES[x][y] is defined, and in fact 
this mapping contains the function that can be used to do the upgrade!
 
NOTE: upgrades of steps should form a linear graph of upgrades to the most up to date
version. For example, if we change add_column from version 1 to version 2 to version 3, 
this object should contain:
    {
        'add_column': {
            1: upgrade_add_column_1_to_add_column_2,
            2: upgrade_add_column_2_to_add_column_3
        }
    }
"""
STEP_UPGRADES_FUNCTION_MAPPING_OLD_FORMAT = {
    'group': {
        1: upgrade_group_1_to_pivot_2
    }, 
    'pivot': {
        2: upgrade_pivot_2_to_pivot_3
    },
    'add_column': {
        1: upgrade_add_column_1_to_add_column_2
    },
    'filter_column': {
        1: update_filter_column_1_to_filter_column_2
    }
}

STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT = {
    'simple_import': {
        1: upgrade_simple_import_1_to_2_and_rename
    },
    'merge': {
        1: upgrade_merge_1_to_merge_2,
        2: upgrade_merge_2_to_3,
        3: upgrade_merge_3_to_4,
    },
    'change_column_dtype': {
        1: upgrade_change_column_dtype_1_to_2,
        2: upgrade_change_column_dtype_2_to_3
    },
    'delete_column': {
        1: upgrade_delete_column_1_to_2,
        2: upgrade_delete_column_2_to_3
    },
    'filter_column': {
        2: upgrade_filter_column_2_to_3,
        3: upgrade_filter_column_3_to_4
    },
    'pivot': {
        3: upgrade_pivot_3_to_4,
        4: upgrade_pivot_4_to_5_and_rename,
        5: upgrade_pivot_5_to_6,
        6: upgrade_pivot_6_to_7,
        7: upgrade_pivot_7_to_8
    },
    'rename_column': {
        1: upgrade_rename_column_1_to_2
    },
    'reorder_column': {
        1: upgrade_reorder_column_1_to_2
    },
    'set_column_formula': {
        1: upgrade_set_column_formula_1_to_2,
        2: upgrade_set_column_formula_2_to_3,
        3: upgrade_set_column_formula_3_to_4
    },
    'sort': {
        1: upgrade_sort_1_to_2
    },
    'graph': {
        1: upgrade_graph_1_to_2,
        2: upgrade_graph_2_to_3,
        3: upgrade_graph_3_to_4
    },
    'change_column_format': {
        1: upgrade_change_column_format_1_to_remove
    },
    'set_dataframe_format': {
        1: upgrade_set_dataframe_format_1_to_2
    }
}


def should_add_initial_bulk_old_rename_step(version: str) -> bool:
    """
    This utility returns True if upgrading should add an initial rename
    step to replace the preprocessing rename step that used to exist at
    the start of the analysis.

    This has to check different versions for mitosheet and mitosheet3,
    to make sure to returns Ture in the correct conditions.
    """

    if package_name == 'mitosheet':
        if is_prev_version(version, '0.1.341'):
            return True
    elif package_name == 'mitosheet3':
        if is_prev_version(version, '0.3.130'):
            return True
    elif package_name == 'mitosheet2':
        # Because mitosheet2 was introduced so much later than this upgrade,
        # we always return false here
        return False

    return False



def upgrade_step_list_to_current(
        step_list: List[Dict[str, Any]], 
        step_upgrade_function_mapping: Dict[str, Dict[int, Callable]]
    ) -> List[Dict[str, Any]]:
    """
    This function upgrades the step list to the most up-to-date format
    for each step, while preserving certain consistency constraints
    while upgrading. 

    It does so by looping over each step, one-by-one, and upgrading it
    repeatedly until it is entirely up to date. Notably, when upgrading 
    each step, it also potentially upgrades all of the steps that come
    after it.

    This makes it possible for us to make upgrades that take place across
    mulitple steps (e.g. the column header -> column id move) to stay
    backwards compatible!

    Finially, as each step is fully upgraded before the next step is 
    processed, this linear approach makes sure that steps are upgraded
    in the correct way. This stops later steps from being changed in
    a way that earlier steps do not expect when they are trying to 
    upgrade them.
    """

    step_index = 0
    upgraded_step_list = copy(step_list)

    while step_index < len(upgraded_step_list):
        step = upgraded_step_list[step_index]

        step_version = step['step_version']
        step_type = step['step_type']
        if step_type in step_upgrade_function_mapping and step_version in step_upgrade_function_mapping[step_type]:
            # Get the upgrade function
            upgrade_function = step_upgrade_function_mapping[step_type][step_version]
            # Upgrade the step, as well as the steps after it
            upgraded_step_and_later_steps = upgrade_function(
                step, 
                # Pass the later steps
                upgraded_step_list[step_index + 1:]
            )
            # Switch the current step and the steps after it with the 
            # newly upgraded ones, and continue trying to upgrade this step
            upgraded_step_list = upgraded_step_list[:step_index]
            upgraded_step_list += upgraded_step_and_later_steps
        else:
            # If the step is up to date, go to the next one to try
            # to upgrade that
            step_index += 1

        
    return upgraded_step_list


def upgrade_steps_for_old_format(saved_analysis: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    """
    A helper function that operates on the old analysis format of:
    {
        "version": "0.1.197",
        "steps": {
            1: ...
        }
    }

    And makes sure all their steps are up to date.
    """
    if saved_analysis is None:
        return None

    # If it's already in the new format, then don't worry about it
    if 'steps' not in saved_analysis and 'steps_data' in saved_analysis:
        return saved_analysis
        
    version = saved_analysis["version"]
    steps = saved_analysis["steps"]

    old_steps = []

    for step_idx, step in steps.items():
        old_steps.append(
            step
        )
    new_steps = upgrade_step_list_to_current(old_steps, STEP_UPGRADES_FUNCTION_MAPPING_OLD_FORMAT)

    # Convert the new steps in the correct format
    new_steps_json = {
        str(i + 1): step for i, step in enumerate(new_steps)
    }

    return {
        'version': version,
        'steps': new_steps_json
    }

def upgrade_steps_for_new_format(saved_analysis: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    """
    A helper function that operates on the new analysis format of:

    "version": "0.1.197",
    "steps_data": [
        {
            "step_version": 1,
            "step_type": "filter_column",
            "params": {
                ....
            }
        }
    ]

    And makes sure all their steps are up to date.
    """
    if saved_analysis is None:
        return None
    
        
    version = saved_analysis["version"]
    steps_data = saved_analysis["steps_data"]

    new_steps_data = []

    # If we should add an initial rename step, we add it at the start of the analysis,
    # to take the place of the preprocessing rename step that used to exist
    if should_add_initial_bulk_old_rename_step(version):
        new_steps_data.append(INITIAL_BULK_OLD_RENAME_STEP)

    new_steps_data.extend(
        upgrade_step_list_to_current(
            steps_data, 
            STEP_UPGRADES_FUNCTION_MAPPING_NEW_FORMAT
        )
    )
    
    return {
        'version': __version__,
        'steps_data': new_steps_data
    }

def upgrade_saved_analysis_to_current_version(saved_analysis: Optional[Dict[str, Any]]) -> Optional[Dict[str, Any]]:
    """
    Upgrades a saved analysis to the current version.
    
    Notable, changes to the saved analysis take two types:
    1. Changes to the format of the saved analysis itself. 
    2. Changes to the format of the specific steps in the saved analysis.

    See mitosheet/upgrade/schemas.py, but as we only have 1 format change
    in the saved analyses, we process the specific step upgrades first if
    they exist.
    """
    saved_analysis = upgrade_steps_for_old_format(saved_analysis)
    new_format_saved_analysis = upgrade_saved_analysis_format_to_steps_data(saved_analysis)
    saved_analysis = upgrade_steps_for_new_format(new_format_saved_analysis)

    return saved_analysis
    


#!/usr/bin/env python
# coding: utf-8

import warnings
# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from time import perf_counter
from typing import Any, Callable, Collection, Dict, List, Optional, Set, Tuple

import pandas as pd

from mitosheet.code_chunks.code_chunk import CodeChunk
from mitosheet.code_chunks.step_performers.pivot_code_chunk import (
    USE_INPLACE_PIVOT, PivotCodeChunk)
from mitosheet.column_headers import flatten_column_header
from mitosheet.errors import make_invalid_pivot_error, make_no_column_error
from mitosheet.state import DATAFRAME_SOURCE_PIVOTED, State
from mitosheet.step_performers.filter import (combine_filters,
                                              get_applied_filter)
from mitosheet.step_performers.step_performer import StepPerformer
from mitosheet.step_performers.utils import get_param
from mitosheet.telemetry.telemetry_utils import log
from mitosheet.types import (ColumnHeader, ColumnHeaderWithFilter,
                             ColumnHeaderWithPivotTransform, ColumnID,
                             ColumnIDWithFilter, ColumnIDWithPivotTransform)
from mitosheet.array_utils import deduplicate_array
from mitosheet.utils import is_prev_version



# Aggregation types pivot supports
PA_COUNT_UNIQUE = 'count unique'
PIVOT_AGGREGATION_TYPES = [
    # These first few are supported out of the box by 
    # pandas, so we don't need any extra support for them
    'sum',
    'mean',
    'median',
    'min',
    'max', 
    'count', 
    'std',
    PA_COUNT_UNIQUE
]

# Pivot Column Transformations: if the user does not specify a transformation before the pivot (e.g. 
# getting the months from a date), then we call this a no-op
PCT_NO_OP = 'no-op'
PCT_DATE_YEAR = 'year'
PCT_DATE_QUARTER = 'quarter'
PCT_DATE_MONTH = 'month'
PCT_DATE_WEEK = 'week'
PCT_DATE_DAY_OF_MONTH = 'day of month'
PCT_DATE_DAY_OF_WEEK = 'day of week'
PCT_DATE_HOUR = 'hour'
PCT_DATE_MINUTE = 'minute'
PCT_DATE_SECOND = 'second'
PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE = 'year-month-day-hour-minute'
PCT_DATE_YEAR_MONTH_DAY_HOUR = 'year-month-day-hour'
PCT_DATE_YEAR_MONTH_DAY = 'year-month-day'
PCT_DATE_YEAR_MONTH = 'year-month'
PCT_DATE_YEAR_QUARTER = 'year-quarter'
PCT_DATE_MONTH_DAY = 'month-day'
PCT_DATE_DAY_HOUR = 'day-hour'
PCT_DATE_HOUR_MINUTE = 'hour-minute'


class PivotStepPerformer(StepPerformer):
    """
    A pivot, which allows you to pivot data from an existing dataframe 
    along some keys, and then aggregate columns with specific functions.
    """

    @classmethod
    def step_version(cls) -> int:
        return 8

    @classmethod
    def step_type(cls) -> str:
        return 'pivot'

    @classmethod
    def saturate(cls, prev_state: State, params: Dict[str, Any]) -> Dict[str, Any]:
        """
        We filter out any duplicated aggregation keys, as they
        result in errors without adding any data to the pivot.
        """
        # Filter out any duplicate aggregation functions
        for column_id, aggregation_function_names in params['values_column_ids_map'].items():
            new_aggregation_function_names = []
            for i in aggregation_function_names:
                if i not in new_aggregation_function_names:
                    new_aggregation_function_names.append(i)
            params['values_column_ids_map'][column_id] = new_aggregation_function_names

        return params

    @classmethod
    def execute(cls, prev_state: State, params: Dict[str, Any]) -> Tuple[State, Optional[Dict[str, Any]]]:
        sheet_index: int = get_param(params, 'sheet_index')
        pivot_rows_column_ids_with_transforms: List[ColumnIDWithPivotTransform] = get_param(params, 'pivot_rows_column_ids_with_transforms')
        pivot_columns_column_ids_with_transforms: List[ColumnIDWithPivotTransform] = get_param(params, 'pivot_columns_column_ids_with_transforms')
        values_column_ids_map: Dict[ColumnID, Collection[str]] = get_param(params, 'values_column_ids_map')
        pivot_filters_ids: List[ColumnIDWithFilter] = get_param(params, 'pivot_filters')
        flatten_column_headers: bool = get_param(params, 'flatten_column_headers')
        destination_sheet_index: Optional[int] = get_param(params, 'destination_sheet_index')
        use_deprecated_id_algorithm: bool = get_param(params, 'use_deprecated_id_algorithm') if get_param(params, 'use_deprecated_id_algorithm') else False
        
        # Get just the column headers in a list, for convenience
        pivot_rows = [prev_state.column_ids.get_column_header_by_id(sheet_index, cit['column_id']) for cit in pivot_rows_column_ids_with_transforms]
        pivot_columns = [prev_state.column_ids.get_column_header_by_id(sheet_index, cit['column_id']) for cit in pivot_columns_column_ids_with_transforms]

        # Make new objects with all columns headers
        pivot_rows_with_transforms: List[ColumnHeaderWithPivotTransform] = [{
            'column_header': prev_state.column_ids.get_column_header_by_id(sheet_index, chwpt['column_id']),
            'transformation': chwpt['transformation']
        } for chwpt in pivot_rows_column_ids_with_transforms]
        pivot_columns_with_transforms: List[ColumnHeaderWithPivotTransform] = [{
            'column_header': prev_state.column_ids.get_column_header_by_id(sheet_index, chwpt['column_id']),
            'transformation': chwpt['transformation']
        } for chwpt in pivot_columns_column_ids_with_transforms]
        values = {
            prev_state.column_ids.get_column_header_by_id(sheet_index, column_id): value 
            for column_id, value in values_column_ids_map.items()
        }
        pivot_filters: List[ColumnHeaderWithFilter] = [
            {'column_header': prev_state.column_ids.get_column_header_by_id(sheet_index, pf['column_id']), 'filter': pf['filter']}
            for pf in pivot_filters_ids
        ]

        # We check that the pivot by doesn't use any columns that don't exist
        columns_used = set(pivot_rows).union(set(pivot_columns)).union(set(values.keys()))
        missing_pivot_keys = columns_used.difference(prev_state.dfs[sheet_index].keys())
        if len(missing_pivot_keys) > 0:
            raise make_no_column_error(missing_pivot_keys)

        # Create the post state, it can be a shallow copy
        post_state = prev_state.copy()

        # Actually execute the pivoting
        pandas_start_time = perf_counter()
        new_df, was_series = _execute_pivot(
            prev_state.dfs[sheet_index], 
            pivot_rows_with_transforms,
            pivot_columns_with_transforms,
            values,
            pivot_filters,
            flatten_column_headers
        )
        pandas_processing_time = perf_counter() - pandas_start_time
        # Create a new df name if we don't have one
        df_name = get_new_pivot_df_name(post_state, sheet_index) if destination_sheet_index is None else post_state.df_names[destination_sheet_index]
        
        destination_sheet_index = post_state.add_df_to_state(
            new_df, 
            DATAFRAME_SOURCE_PIVOTED,
            sheet_index=destination_sheet_index,
            df_name=df_name,
            use_deprecated_id_algorithm=use_deprecated_id_algorithm
        )


        return post_state, {
            'destination_sheet_index': destination_sheet_index,
            'was_series': was_series,
            'pandas_processing_time': pandas_processing_time
        }

    @classmethod
    def transpile(
        cls,
        prev_state: State,
        post_state: State,
        params: Dict[str, Any],
        execution_data: Optional[Dict[str, Any]],
    ) -> List[CodeChunk]:
        return [
            PivotCodeChunk(
                prev_state, 
                post_state, 
                get_param(params, 'sheet_index'),
                get_param(params, 'destination_sheet_index'),
                get_param(params, 'pivot_rows_column_ids_with_transforms'),
                get_param(params, 'pivot_columns_column_ids_with_transforms'),
                get_param(params, 'pivot_filters'),
                get_param(params, 'values_column_ids_map'),
                get_param(params, 'flatten_column_headers'),
                get_param(execution_data if execution_data is not None else {}, 'was_series'),
            )
        ]

    @classmethod
    def get_modified_dataframe_indexes(cls, params: Dict[str, Any]) -> Set[int]:
        destination_sheet_index = get_param(params, 'destination_sheet_index')
        if destination_sheet_index: # If editing an existing sheet, that is what is changed
            return {destination_sheet_index}
        return {-1}
    
def values_to_functions(values: Dict[ColumnHeader, Collection[str]]) -> Dict[ColumnHeader, List[Callable]]:
    """
    Helper function for turning the values mapping sent by the frontend to 
    the value map of functions that can actually be passed to the pandas pivot function
    """
    new_values: Dict[ColumnHeader, List[Callable]] = dict()

    for column_header, aggregation_function_names in values.items():
        new_values[column_header] = []
        for agg_func_name in aggregation_function_names:
            if agg_func_name not in PIVOT_AGGREGATION_TYPES:
                raise make_invalid_pivot_error()

            if agg_func_name == PA_COUNT_UNIQUE:
                agg_func = pd.Series.nunique
            else:
                agg_func = agg_func_name
                
            new_values[column_header].append(agg_func)
    
    return new_values


def get_new_column_header_from_column_header_with_pivot_transform(chwpt: ColumnHeaderWithPivotTransform) -> ColumnHeader:
    column_header, transformation = chwpt['column_header'], chwpt['transformation']
    # We need to turn the column header into a string before creating the new one, so that we can
    # append to it for the new temporary transformation column
    if transformation == PCT_NO_OP:
        return column_header

    return f'{str(column_header)} ({transformation})'

def add_transform_columns_to_dataframe(df: pd.DataFrame, column_headers_with_transforms: List[ColumnHeaderWithPivotTransform]) -> pd.DataFrame:
    for chwpt in column_headers_with_transforms:
        column_header, transformation = chwpt['column_header'], chwpt['transformation']
        # We need to turn the column header into a string before creating the new one, so that we can
        # append to it for the new temporary transformation column
        if transformation == PCT_NO_OP:
            continue

        new_column_header = get_new_column_header_from_column_header_with_pivot_transform(chwpt)
        if transformation == PCT_DATE_YEAR:
            df[new_column_header] = df[column_header].dt.year
        if transformation == PCT_DATE_QUARTER:
            df[new_column_header] = df[column_header].dt.quarter
        if transformation == PCT_DATE_MONTH:
            df[new_column_header] = df[column_header].dt.month
        if transformation == PCT_DATE_WEEK:
            if is_prev_version(pd.__version__, '1.0.0'):
                df[new_column_header] = df[column_header].dt.week
            else:
                df[new_column_header] = df[column_header].dt.isocalendar().week.astype(int)
        if transformation == PCT_DATE_DAY_OF_MONTH:
            df[new_column_header] = df[column_header].dt.day
        if transformation == PCT_DATE_DAY_OF_WEEK:
            df[new_column_header] = df[column_header].dt.weekday
        if transformation == PCT_DATE_HOUR:
            df[new_column_header] = df[column_header].dt.hour
        if transformation == PCT_DATE_MINUTE:
            df[new_column_header] = df[column_header].dt.minute
        if transformation == PCT_DATE_SECOND:
            df[new_column_header] = df[column_header].dt.second
        if transformation == PCT_DATE_YEAR_MONTH_DAY_HOUR_MINUTE:
            df[new_column_header] = df[column_header].dt.strftime("%Y-%m-%d %H:%M")
        if transformation == PCT_DATE_YEAR_MONTH_DAY_HOUR:
            df[new_column_header] = df[column_header].dt.strftime("%Y-%m-%d %H")
        if transformation == PCT_DATE_YEAR_MONTH_DAY:
            df[new_column_header] = df[column_header].dt.strftime("%Y-%m-%d")
        if transformation == PCT_DATE_YEAR_MONTH:
            df[new_column_header] = df[column_header].dt.strftime("%Y-%m")
        if transformation == PCT_DATE_YEAR_QUARTER:
            df[new_column_header] = df[column_header].dt.year.astype(str) + "-Q" + df[column_header].dt.quarter.astype(str)
        if transformation == PCT_DATE_MONTH_DAY:
            df[new_column_header] = df[column_header].dt.strftime("%m-%d")
        if transformation == PCT_DATE_DAY_HOUR:
            df[new_column_header] = df[column_header].dt.strftime("%d %H")
        if transformation == PCT_DATE_HOUR_MINUTE:
            df[new_column_header] = df[column_header].dt.strftime("%H:%M")

    
    return df


def _execute_pivot(
        df: pd.DataFrame, 
        pivot_rows_with_transforms: List[ColumnHeaderWithPivotTransform], 
        pivot_columns_with_transforms: List[ColumnHeaderWithPivotTransform], 
        values: Dict[ColumnHeader, Collection[str]],
        pivot_filters: List[ColumnHeaderWithFilter],
        flatten_column_headers: bool
    ) -> Tuple[pd.DataFrame, bool]:
    """
    Helper function for executing the pivot on a specific dataframe
    and then aggregating the values with the passed values mapping
    """

    pivot_rows = [cit['column_header'] for cit in pivot_rows_with_transforms]
    pivot_columns = [cit['column_header'] for cit in pivot_columns_with_transforms]

    # If there are no keys to aggregate on, we return an empty dataframe
    if (len(pivot_rows) == 0 and len(pivot_columns) == 0) or len(values) == 0:
        return pd.DataFrame(data={}), False

    values_keys = list(values.keys())

    # First, we do the filtering on the initial dataframe, according to the pivot_filters
    if len(pivot_filters) > 0:
        filters = [
            get_applied_filter(df, pf['column_header'], pf['filter'])
            for pf in pivot_filters
        ]
        full_filter = combine_filters('And', filters)
        df = df[full_filter] # TODO: do we have to make a copy

    # Then, we make a temp dataframe that does not have the columns 
    # we do not need, as this allows us to avoid a bug in pandas where these extra
    # columns cause a data
    unused_columns = df.columns.difference(set(pivot_rows).union(set(pivot_columns)).union(set(values_keys)))
    # NOTE: We need to make a copy, so that we can set columns on a full dataframe rather than
    # just a view of a different dataframe. This is just a shallow copy
    df = df.drop(unused_columns, axis=1).copy()

    # Then, we create the new columns that are a function of the transforms that are passed
    # with the pivot params
    df = add_transform_columns_to_dataframe(df, pivot_rows_with_transforms)
    df = add_transform_columns_to_dataframe(df, pivot_columns_with_transforms)

    # Create the final pivot_rows and pivot_columns, which might be the temporary columns
    # created above by the transformations. NOTE: we need to deduplicate these here, as
    # having multiple rows or columns results in errors in the pandas pivot function
    final_pivot_rows = deduplicate_array([
        get_new_column_header_from_column_header_with_pivot_transform(chwpt)
        for chwpt in pivot_rows_with_transforms
    ])
    final_pivot_columns = deduplicate_array([
        get_new_column_header_from_column_header_with_pivot_transform(chwpt)
        for chwpt in pivot_columns_with_transforms
    ])

    # Built the args, leaving out any unused values
    args: Dict[str, Any] = {}
    if len(final_pivot_rows) > 0:
        args['index'] = final_pivot_rows

    if len(final_pivot_columns) > 0:
        args['columns'] = final_pivot_columns

    if len(values) > 0:
        args['values'] = values_keys
        args['aggfunc'] = values_to_functions(values)
        

    # While performing the pivot table, catch warnings that are created
    # by pandas so that we can log them.
    with warnings.catch_warnings(record=True):
        # Forward the warning handling to our custom function to log it
        warnings.showwarning = log_pivot_table_warnings
        
        # Actually perform the pivot
        pivot_table = df.pivot_table(**args) # type: pd.DataFrame

    # On earlier pandas versions (e.g. 0.24.2), the pivot table function returned
    # a series from the above function call. Thus, we need to move it to a df for
    # all our other code run properly on it. This code should only run in early 
    # versions of pandas
    was_series = False
    if isinstance(pivot_table, pd.Series):
        pivot_table = pd.DataFrame(pivot_table)
        was_series = True

    if flatten_column_headers:
        # See comment in pivot_code_chunk, we avoid using inplace=True post pandas 1.5.0, as
        # it is depricated
        if USE_INPLACE_PIVOT:
            pivot_table.set_axis([flatten_column_header(col) for col in pivot_table.keys()], axis=1, inplace=True)
        else:
            pivot_table = pivot_table.set_axis([flatten_column_header(col) for col in pivot_table.keys()], axis=1)

    # Reset the indexes of the pivot table
    pivot_table = pivot_table.reset_index()

    return pivot_table, was_series

def get_new_pivot_df_name(post_state: State, sheet_index: int) -> str: 
    """
    Creates the name for the new pivot table sheet using the format
    {source_sheet_name}_{pivot} or {source_sheet_name}_{pivot}_1, etc. 
    if the pivot table name already exists. 
    """
    new_df_name_original = post_state.df_names[sheet_index] + '_pivot'
    curr_df_name = new_df_name_original
    multiple_sheet_indicator = 1
    while curr_df_name in post_state.df_names:
        curr_df_name = f'{new_df_name_original}_{multiple_sheet_indicator}'
        multiple_sheet_indicator += 1
    return curr_df_name


def log_pivot_table_warnings(message, *args):
    """
    Logs warnings that are created by the pandas pivot table function.
    Using warnings.showwarning passes several additional arguments to the function, 
    so we must include *args to support them.
    """
    params = {'message': str(message)}
    log('pivot_table_performance_warning', params)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Saga Inc.
# Distributed under the terms of the GPL License.
from typing import Any, Dict, List
import warnings
import pytest
import pandas as pd

from mitosheet.errors import MitoError
from mitosheet.parser import get_backend_formula_from_frontend_formula, parse_formula, safe_replace, safe_contains, get_frontend_formula
from mitosheet.types import FORMULA_ENTIRE_COLUMN_TYPE


def get_number_data_for_df(columns: List[Any], length: int) -> Dict[Any, Any]:
    return {ch: [i for i in range(length)] for ch in columns}


def get_string_data_for_df(columns: List[Any], length: int) -> Dict[Any, Any]:
    return {ch: [str(i) for i in range(length)] for ch in columns}

CONSTANT_TEST_CASES: Any = [
    (
        '=100',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = 100',
        set([]),
        set([])
    ),
    (
        '=True',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = True',
        set([]),
        set([])
    ),
    (
        '=\'StringSingleQuotes\'',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \'StringSingleQuotes\'',
        set([]),
        set([])
    ),
    (
        '=\"StringDoubleQuotes\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"StringDoubleQuotes\"',
        set([]),
        set([])
    ), 
    (
        '=\"String Double Quotes\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"String Double Quotes\"',
        set([]),
        set([])
    ), 
    (
        '=FUNC(\"String Double Quotes\")',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC(\"String Double Quotes\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=FUNC(\"DIFF_FUNC(A)\")',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC(\"DIFF_FUNC(A)\")',
        set(['FUNC']),
        set([])
    ),
    (
        '=\"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \"String One FUNC(A)\" + \"STRING TWO FUNC(A)\"',
        set([]),
        set([])
    ),
    (
        '=\'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = \'String One FUNC(A)\' + \'STRING TWO FUNC(A)\'',
        set([]),
        set([])
    ),
    (
        '=FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        'B',
        0,
        pd.DataFrame(get_string_data_for_df(['B'], 2)),
        'df[\'B\'] = FUNC1(\'String One FUNC(A)\') + FUNC2(\'STRING TWO FUNC(A)\')',
        set(['FUNC1', 'FUNC2']),
        set([])
    )
]

# Tests cases to ensure operators are parsed correctly
OPERATOR_TEST_CASES = [
    # Simple add operation
    (
        '=A + B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] + df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple subtract operation
    (
        '=A - B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] - df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple multiply operation
    (
        '=A * B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] * df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Simple divide operation
    (
        '=A / B',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = df[\'A\'] / df[\'B\']',
        set(),
        set(['A', 'B'])
    ),
    # Handles multi-char columns
    (
        '=AAA + BBB + CCC + DDD',
        'E',
        0,
        pd.DataFrame(get_number_data_for_df(['AAA', 'BBB', 'CCC', 'DDD'], 2)),
        'df[\'E\'] = df[\'AAA\'] + df[\'BBB\'] + df[\'CCC\'] + df[\'DDD\']',
        set(),
        set(['AAA', 'BBB', 'CCC', 'DDD'])
    ),
    # Maintains parens and constants
    (
        '=(A + B) / C + A * 100',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = (df[\'A\'] + df[\'B\']) / df[\'C\'] + df[\'A\'] * 100',
        set(),
        set(['A', 'B', 'C'])
    ),
    # Operator in functions
    (
        '=FUNC(A + B / C + A * 100)',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = FUNC(df[\'A\'] + df[\'B\'] / df[\'C\'] + df[\'A\'] * 100)',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    )
]

# Tests proper function parsing
FUNCTION_TEST_CASES = [
    # Simple, singular function call
    (
        '=FUNC(A)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Simple, singular function call that has the name of a column
    (
        '=C(A)',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = C(df[\'A\'])',
        set(['C']),
        set(['A'])
    ),
    # Nested functions with different names
    (
        '=FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(A))))))))',
        'B',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'B\'] = FUNCA(FUNCB(FUNCC(FUNCD(FUNCE(FUNCF(FUNCG(FUNCH(df[\'A\']))))))))',
        set(['FUNCA', 'FUNCB', 'FUNCC', 'FUNCD', 'FUNCE', 'FUNCF', 'FUNCG', 'FUNCH']),
        set(['A'])
    ),
    # Nested functions with basic operators
    (
        '=RIGHT(A, LEN(B) - 10)',
        'C',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'C\'] = RIGHT(df[\'A\'], LEN(df[\'B\']) - 10)',
        set(['RIGHT', 'LEN']),
        set(['A', 'B'])
    ),
    # More nested functions with columns
    (
        '=FUNC(A, FUNC(B, FUNC(C)))',
        'D',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C'], 2)),
        'df[\'D\'] = FUNC(df[\'A\'], FUNC(df[\'B\'], FUNC(df[\'C\'])))',
        set(['FUNC']),
        set(['A', 'B', 'C'])
    ),
    # Lots of column references
    (
        '=FUNC(A, B, C, D, E, F, G, H, I)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'], 2)),
        'df[\'J\'] = FUNC(df[\'A\'], df[\'B\'], df[\'C\'], df[\'D\'], df[\'E\'], df[\'F\'], df[\'G\'], df[\'H\'], df[\'I\'])',
        set(['FUNC']),
        set(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I'])
    ),
    # Lots of column references to columns with numbers
    (
        '=FUNC(column1, column2)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['column1', 'column2', 'J'], 2)),
        'df[\'J\'] = FUNC(df[\'column1\'], df[\'column2\'])',
        set(['FUNC']),
        set(['column1', 'column2'])
    ),
    # Test references to columns with spaces
    (
        '=FUNC(column 1, column 2)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['column 1', 'column 2', 'J'], 2)),
        'df[\'J\'] = FUNC(df[\'column 1\'], df[\'column 2\'])',
        set(['FUNC']),
        set(['column 1', 'column 2'])
    ),
    # Test references to columns that are numbers
    (
        '=FUNC(123, 456)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df([123, 456, 'J'], 2)),
        'df[\'J\'] = FUNC(df[123], df[456])',
        set(['FUNC']),
        set([123, 456])
    ),
    # Test references to columns that are numbers and are also strings
    (
        '=FUNC(123, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 123, "J"], 2)),
        'df[\'J\'] = FUNC(df[123], df[456], df[\'abc\'])',
        set(['FUNC']),
        set([123, 456, 'abc'])
    ),
    # Test references to strings that has a single quotes in it
    (
        '=FUNC(this is a string \' that has a quote, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 'this is a string \' that has a quote', "J"], 2)),
        'df[\'J\'] = FUNC(df[\'this is a string \' that has a quote\'], df[456], df[\'abc\'])',
        set(['FUNC']),
        set(['abc', 456, 'this is a string \' that has a quote'])
    ),
    # Test references to strings that have mulitple quotes
    (
        '=FUNC(this is a string \'test\' that has a quote, 456, abc)',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(['abc', 456, 'this is a string \'test\' that has a quote', "J"], 2)),
        'df[\'J\'] = FUNC(df[\'this is a string \'test\' that has a quote\'], df[456], df[\'abc\'])',
        set(['FUNC']),
        set(['abc', 456, 'this is a string \'test\' that has a quote'])
    ),
    # Test references to a string in quotes, but is a column header
    (
        '=FUNC("column")',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(["\"column\"", "J"], 2)),
        'df[\'J\'] = FUNC(df[\'"column"\'])',
        set(['FUNC']),
        set(["\"column\""])
    ),
    # Test references to a quote that is covered by a column header
    (
        '=FUNC("this is a string")',
        'J',
        0,
        pd.DataFrame(get_number_data_for_df(["string", "J"], 2)),
        'df[\'J\'] = FUNC("this is a string")',
        set(['FUNC']),
        set([])
    ),
    # Test references a boolean column
    (
        '=true + false',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df([True, False], 2)),
        'df[\'A\'] = df[True] + df[False]',
        set([]),
        set([True, False])
    ),
    # Test references a multi-index header
    (
        '=Height, min',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = df[(\'Height\', \'min\')]',
        set([]),
        set([('Height', 'min')])
    ),
    # Test references a multi-index header in a function
    (
        '=FUNC(Height, min)',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = FUNC(df[(\'Height\', \'min\')])',
        set(['FUNC']),
        set([('Height', 'min')])
    ),
    # Test references a multi-index header with an empty second item
    (
        '=FUNC(Name)',
        'Name',
        0,
        pd.DataFrame(get_number_data_for_df([('Name', ''), ('Height', 'min')], 2)),
        'df[\'Name\'] = FUNC(df[(\'Name\', \'\')])',
        set(['FUNC']),
        set([('Name', '')])
    ),
    # Test references a header with characters that make an invalid regex
    (
        '=FUNC(C++)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'C++'], 2)),
        'df[\'A\'] = FUNC(df[\'C++\'])',
        set(['FUNC']),
        set(['C++'])
    ),
]

# To supress warnings, we create these variables here
with warnings.catch_warnings():
    warnings.simplefilter('ignore')
    unit64_index = pd.UInt64Index(range(10))
    float64_index = pd.Float64Index(range(10))


INDEX_TEST_CASES = [
    # Test references a range index with a valid number 0
    (
        'A0',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test references a range index with a valid number 0, with two
    (
        'A0 + A0',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = df[\'A\'] + df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test references a range index with a valid number 0, function
    (
        '=FUNC(A0)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a valid number non 0
    (
        '=FUNC(A0, A1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a valid number non 0, positive shift
    (
        '=FUNC(A0, A1)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index with a number that is NOT part of the index
    (
        '=FUNC(A0, A2)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], A2)',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index that does not start at 0 for proper offset
    (
        '=FUNC(A10, A11)',
        'A',
        10,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(10, 12)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index that does not start at 0 for proper offset, positive shift
    (
        '=FUNC(A10, A11)',
        'A',
        11,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(10, 12)),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a different number index
    (
        '=FUNC(A0, A2)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 10), index=unit64_index),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a number dtype
    (
        '=FUNC(A0.0, A2.2)',
        'A',
        0.0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0.0, 1.1, 2.2], dtype='float64')),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a number dtype, positive shift
    (
        '=FUNC(A0.0, A2.2)',
        'A',
        2.2,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0.0, 1.1, 2.2], dtype='float64')),
        'df[\'A\'] = FUNC(df[\'A\'].shift(2, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from 0
    (
        '=FUNC(A1)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.RangeIndex(1, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from below 0
    (
        '=FUNC(A-2)',
        'A',
        -2,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 4), index=pd.RangeIndex(-2, 2)),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index not starting from below 0, positive shift
    (
        '=FUNC(A1, A-2)',
        'A',
        1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 4), index=pd.RangeIndex(-2, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(3, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a range index starting with a decimal
    (
        '=FUNC(A0.1, A0.2)',
        'A',
        .1,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([.1, .2, .3])),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference no offset
    (
        '=FUNC(A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 day
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 day, positive shift
    (
        '=FUNC(A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-23 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with offset of 1 month, should still only have offset of 1
    (
        '=FUNC(A2007-02-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-02-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime reference with no offset, and some offset
    (
        '=FUNC(A2007-02-22 00:00:00, A2007-01-22 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-02-22 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0), df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with no offset
    (
        '=FUNC(Aa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.Index(['a'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with at the end of the formula
    (
        '=Aa',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 1), index=pd.Index(['a'])),
        'df[\'A\'] = df[\'A\']',
        set([]),
        set(['A'])
    ),
    # Test a string index with 1 offset
    (
        '=FUNC(Ab)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with with 1 offset, positive
    (
        '=FUNC(Aa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index with spaces
    (
        '=FUNC(Athis has spaces)',
        'A',
        'this has spaces',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['this has spaces', 'this also has spaces'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that contains a column header, no shift
    (
        '=FUNC(AthisA)',
        'A',
        'thisA',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['thisA', 'also this'])),
        'df[\'A\'] = FUNC(df[\'A\'])',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that contains a column header, shift
    (
        '=FUNC(AthisA)',
        'A',
        'also this',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['also this', 'thisA'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that has prefixes
    (
        '=FUNC(Aaa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 3), index=pd.Index(['b', 'a', 'aa'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a string index that has prefixes with spaces
    (
        '=FUNC(Aa aa)',
        'A',
        'b',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 3), index=pd.Index(['b', 'a ', 'a aa'])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a datetime column has no fill value
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame({'A': [pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')]}, index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test a boolean column has no fill value
    (
        '=FUNC(A2007-01-23 00:00:00)',
        'A',
        pd.to_datetime('2007-01-22 00:00:00'),
        pd.DataFrame({'A': [True]}, index=pd.DatetimeIndex([pd.to_datetime('2007-01-22 00:00:00'), pd.to_datetime('2007-01-23 00:00:00')])),
        'df[\'A\'] = FUNC(df[\'A\'].shift(-1))',
        set(['FUNC']),
        set(['A'])
    ),
]

INDEX_TEST_CASES_THAT_DONT_RECONSTRUCT_EXACTLY = [
    # Test references a column header and a range specifically
    (
        '=FUNC(A, A1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references a different number index
    (
        '=FUNC(A0.0, A2.0, A7)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 10), index=float64_index),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0), df[\'A\'].shift(-7, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references an index with a int dtype
    (
        '=FUNC(A0.0, A2.0)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['A', 'B'], 3), index=pd.Index([0, 1, 2], dtype='int64')),
        'df[\'A\'] = FUNC(df[\'A\'], df[\'A\'].shift(-2, fill_value=0))',
        set(['FUNC']),
        set(['A'])
    ),
    # Test references to a column header that contains the index, and one that is unqualified
    (
        '=FUNC(HEADER00, HEADER0HEADER1)',
        'A',
        0,
        pd.DataFrame(get_number_data_for_df(['HEADER0', 'HEADER0HEADER1'], 2), index=pd.RangeIndex(0, 2)),
        'df[\'A\'] = FUNC(df[\'HEADER0\'], df[\'HEADER0HEADER1\'])',
        set(['FUNC']),
        set(['HEADER0', 'HEADER0HEADER1'])
    ),
    # Test a string index with column header containing the index
    (
        '=FUNC(aaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'])',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column header containing the index, no shift
    (
        '=FUNC(aaaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'])',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column header containing the index, no shift
    (
        '=FUNC(aaab)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['A', 'aaa'], 2), index=pd.Index(['a', 'b'])),
        'df[\'A\'] = FUNC(df[\'aaa\'].shift(-1, fill_value=0))',
        set(['FUNC']),
        set(['aaa'])
    ),
    # Test a string index with column headers AND indexes that are prefixes, just detects column header
    (
        '=FUNC(aaaa)',
        'A',
        'a',
        pd.DataFrame(get_number_data_for_df(['aaa', 'aaaa'], 2), index=pd.Index(['a', 'aa'])),
        'df[\'A\'] = FUNC(df[\'aaaa\'])',
        set(['FUNC']),
        set(['aaaa'])
    ),
]


"""
PARSE_TESTS contains a variety of tests that make sure
formula parsing is working properly; it is passed as 
a parameter into the test_parse test below.

Order of params is: formula, address, python_code, functions, columns

See documentation here: https://docs.pytest.org/en/latest/parametrize.html#parametrize-basics
"""
PARSE_TESTS = CONSTANT_TEST_CASES + OPERATOR_TEST_CASES + FUNCTION_TEST_CASES + INDEX_TEST_CASES + INDEX_TEST_CASES_THAT_DONT_RECONSTRUCT_EXACTLY
@pytest.mark.parametrize("formula,column_header,formula_label,df,python_code,functions,columns", PARSE_TESTS)
def test_parse(formula, column_header, formula_label, df, python_code, functions, columns):
    code, funcs, cols, _ = parse_formula(formula, column_header, formula_label, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, df) 
    assert (code, funcs, cols) == \
        (
            python_code, 
            functions, 
            columns
        )


PARSE_TEST_ERRORS = [
    ('=LOOKUP(100, A)', 'B', 'invalid_formula_error', 'LOOKUP'),
    ('=VLOOKUP(100, A)', 'B', 'invalid_formula_error', 'VLOOKUP'),
    ('=HLOOKUP(100, A)', 'B', 'invalid_formula_error', 'HLOOKUP'),
    ('=XLOOKUP(100, A)', 'B', 'invalid_formula_error', 'XLOOKUP'),
    ('=A <> 100', 'B', 'invalid_formula_error', '<>'),
    ('=SUM(A', 'B', 'invalid_formula_error', 'parentheses'),
]
@pytest.mark.parametrize("formula, address, error_type, to_fix_substr", PARSE_TEST_ERRORS)
def test_parse_errors(formula, address, error_type, to_fix_substr):
    with pytest.raises(MitoError) as e_info:
        parse_formula(formula, address, 0, {'type': FORMULA_ENTIRE_COLUMN_TYPE}, pd.DataFrame(get_number_data_for_df(['A', 'B'], 2)))
    assert e_info.value.type_ == error_type
    if to_fix_substr is not None:
        assert to_fix_substr in e_info.value.to_fix


SAFE_REPLACE_TESTS = [
    ('=A', 'A', 'B', ['A', 'B'], '=B'),
    ('=A + A', 'A', 'B', ['A', 'B'], '=B + B'),
    ('=A + B', 'A', 'B', ['A', 'B'], '=B + B'),
    ('=A + B + A', 'A', 'B', ['A', 'B'], '=B + B + B'),
    ('=This is a test', 'test', 'B', ['This is a test'], '=This is a test'),
    ('="test with spaces"', 'test with spaces', 'B', ['test with spaces'], '="test with spaces"'),
    ('=A + B + \"A\"', 'A', 'B', ['A', 'B'], '=B + B + \"A\"'),
    ('=A + B + \'A\'', 'A', 'B', ['A', 'B'], '=B + B + \'A\''),
    ('=A + B + \'A\'', 'A', 'B', ['A', 'B'], '=B + B + \'A\''),
    ('=FUNC(A, B, A) + TEST(FUNC(A, \'A\', \'B\')) + \'A\'', 'A', 'B', ['A', 'B'], '=FUNC(B, B, B) + TEST(FUNC(B, \'A\', \'B\')) + \'A\''),
    ('=APPLE(A, B, A) + AARON(AAA(A, \'A\', \'B\')) + \'A\'', 'A', 'B', ['A', 'B'], '=APPLE(B, B, B) + AARON(AAA(B, \'A\', \'B\')) + \'A\''),
]

@pytest.mark.parametrize('formula,old_column_header,new_column_header,column_headers,new_formula', SAFE_REPLACE_TESTS)
def test_safe_replace(formula, old_column_header, new_column_header, column_headers, new_formula):
    assert safe_replace(formula, old_column_header, new_column_header, 0, pd.DataFrame(get_number_data_for_df(column_headers, 2))) == new_formula


SAFE_CONTAINS_TESTS = [
    ('=A & "B"', '&', True),
    ('=A - +  "&"', '&', False),
    ('=A - +  \'&\'', '&', False),
]

@pytest.mark.parametrize('formula,substring,contains', SAFE_CONTAINS_TESTS)
def test_safe_contains(formula, substring, contains):
    assert safe_contains(formula, substring, ['A', 'B']) == contains


@pytest.mark.parametrize("formula,column_header,formula_label,df,python_code,functions,columns", INDEX_TEST_CASES)
def test_get_frontend_formula_reconstucts_properly(formula,column_header,formula_label,df,python_code,functions,columns):
    frontend_formula = get_frontend_formula(formula, formula_label, df)
    assert get_backend_formula_from_frontend_formula(frontend_formula, formula_label, df) == formula



from copy import copy
from typing import List, Tuple
from mitosheet.code_chunks.code_chunk import CodeChunk



def optimize_code_chunks_combine_right(code_chunks_to_optimize: List[CodeChunk]) -> Tuple[bool, List[CodeChunk]]:
    """
    Given a list of code chunks, will attempt to optimize them right
    with a single pass.
    """

    code_chunks_to_optimize = copy(code_chunks_to_optimize)
    code_chunks_to_optimize.reverse() # Reverse the list so we can pop from the old front quickly

    code_chunks_list: List[CodeChunk] = []

    optimized = False
    while len(code_chunks_to_optimize) >= 2:
        later_code_chunk = code_chunks_to_optimize.pop()
        earlier_code_chunk = code_chunks_to_optimize.pop()

        combined_chunk = later_code_chunk.combine_right(earlier_code_chunk)

        if combined_chunk is not None:
            # If we can combine the two chunks into one, take that
            optimized = True
            code_chunks_to_optimize.append(combined_chunk)
        else:
            # If we cannot combine the two chunks, take the first chunk, 
            # and reset the second one to the new first chunk for the next loop
            code_chunks_list.append(later_code_chunk)
            code_chunks_to_optimize.append(earlier_code_chunk)            
    
    # Make sure we take the final item in the code chunks list, 
    # as it has nothing to combine_right with
    if len(code_chunks_to_optimize) == 1:
        code_chunks_list.append(code_chunks_to_optimize[0])

    return optimized, code_chunks_list

def optimize_code_chunks_combine_left(code_chunks_to_optimize: List[CodeChunk]) -> Tuple[bool, List[CodeChunk]]:
    """
    Given a list of code chunks, will attempt to optimize them left
    with a single pass.
    """

    code_chunks_to_optimize = copy(code_chunks_to_optimize)
    code_chunks_list: List[CodeChunk] = []

    optimized = False
    while len(code_chunks_to_optimize) >= 2:
        later_code_chunk = code_chunks_to_optimize.pop()
        earlier_code_chunk = code_chunks_to_optimize.pop()

        combined_chunk = later_code_chunk.combine_left(earlier_code_chunk)

        if combined_chunk is not None:
            # If we can combine the two chunks into one, take that
            optimized = True
            code_chunks_to_optimize.append(combined_chunk)
        else:
            # If we cannot combine the two chunks, take the first chunk, 
            # and reset the second one to the new first chunk for the next loop
            code_chunks_list.append(later_code_chunk)
            code_chunks_to_optimize.append(earlier_code_chunk)            
    
    # Make sure we take the final item in the code chunks list, 
    # as it has nothing to combine_left with
    if len(code_chunks_to_optimize) == 1:
        code_chunks_list.append(code_chunks_to_optimize[0])

    # Since we build the result backwards, we have to reverse it at the end
    code_chunks_list.reverse()

    return optimized, code_chunks_list


def optimize_code_chunks(all_code_chunks: List[CodeChunk]) -> List[CodeChunk]:
    """
    Given a list of code chunks, will recursively attempt to optimize them 
    down to the smallest list of code chunks that have the same effects
    as the original list. 

    This is necessarily recursive, because of a situation like [A, A, B, B], 
    where A and B can be combined to a No-op. Thus, after one call, we end with 
    [A, B], and we need to recurse to finish the optimization.
    """

    optimized_right, code_chunks_list = optimize_code_chunks_combine_right(all_code_chunks)
    optimized_left, code_chunks_list = optimize_code_chunks_combine_left(code_chunks_list)
    
    # As long as we optimized in this iteration, recurse as we might
    # need to optimize again
    if optimized_right or optimized_left:
        return optimize_code_chunks(code_chunks_list)

    return code_chunks_list